"""
Contains the ControllerArchon class.
Originally written by Grzegorz Zareba.
"""

import socket
import time
import threading

import azcam
import azcam.utils
import azcam.exceptions
import azcam.sockets
from azcam.tools.controller import Controller


class ControllerArchon(Controller):
    """
    The controller class for STA Archon controllers.
    """

    # Exposure states
    EXP_UNKNOWN = 0
    EXP_IDLE = 1
    EXP_EXPOSE = 2
    EXP_READY = 3
    EXP_FETCH = 4
    EXP_DONE = 5

    power_values = [
        "UNKNOWN",
        "NOT_CONFIGURED",
        "OFF",
        "INTERMEDIATE",
        "ON",
        "STADBY",
    ]

    def __init__(self, tool_id="controller", description=None):
        super().__init__(tool_id, description)

        self.controller_class = "archon"
        self.controller_type = "archon"

        # do not initialize this controller type
        azcam.db.tools_init.pop("controller", None)

        # azcam connected to the controller
        self.connected_controller = 0

        # parameters
        self.parameters = {}

        self.timing_file = ""

        self.heater_board_installed = 0

        # reset flag
        self.reset_flag = 0  # 0 for soft reset, 1 to upload code

        # status keywords
        self.status_valid = 0
        self.status_count = 0
        self.status_log = 0
        self.status_power = 0
        self.status_backplane_temp = 0.0

        self.cont_exposures = 0
        self.exposures = 1
        self.sweep_cnt = 1
        self.exp_time_ms = 0
        self.int_ms = 0
        self.noint_ms = 0
        self.pixels = 0
        self.lines = 0

        # response to the FRAME command
        self.frame = 0
        # FRAME data dictionary
        self.dict_frame = {}

        # response to the STATUS command
        self.status = 0
        # STATUS data dictionary
        self.dict_status = {}

        # ronfig data lines read from a file or downloaded - raw format 'parameter=value'
        self.config_data = []

        # Sorted Parameters section of the configuration data
        self.config_params = []

        # Config data lines counter
        self.config_lines_cnt = 0

        # Config data lines:  parameter:value pairs
        self.dict_config = {}

        # Config data lines positions: parameter_name:position pairs
        self.dict_wconfig = {}

        # Parameters positions: parameter_name:parameter_position pairs
        self.dict_params = {}

        # Taplines dictionary
        self.dict_taplines = {}

        # Config data OK flag
        self.config_ok = 0

        self.read_buffer = 0

        self.currframe1 = 0
        self.currframe2 = 0
        self.currframe3 = 0

        # Config data read from a file of from the GUI - raw format 'parameter=value'
        self.config_data = ""
        # Config data dictionary

        self.power_status = "UNKNOWN"
        self.archon_status = 0

        self.exp_start = 0
        self.exp_timer = 0
        self.newframe = 0

        # image type
        self.img_type = 0

        # Image data received from the Archom controller in Direct Mode
        self.imagedata = 0

        # video gain flag
        self.video_gain = 1

        # For timing
        self.frame_time = 0

        # not used - fix me
        self.timing_board = ""
        self.clock_boards = [""]
        self.video_boards = [""]

        self.pixels = 0
        self.lines = 0

        # video gain flag
        self.video_gain = 1

        # CDS parameters (taplines) - ex: ["AD1L, -3.0, 3000", "AD2R, 3.0, 3000"]
        self.cds = []

        # CDS parameters read from the Archon controller
        self.rcds = []

        # CDS parameters string for console update
        self.ucds = ""

        # Total number of taplines
        self.tap_lines = 0

        # receiving raw data
        self.rawdata_enable = 0
        # Raw channel - in the config dictionary raw channel numbers start from 0
        self.rawdata_channel = 1
        # Raw data received from the Archon controller
        self.rawdata = 0

        #: lock for threads
        self.lock = threading.Lock()

        # controller server
        self.camserver = azcam.sockets.SocketInterface()
        self.camserver.host = ""
        self.camserver.port = 4242
        self.camserver.cmd_id = 0x00
        self.camserver.lastcmd_id = 0x00

    def connect(self):
        """
        Connects Azcam to the controller.
        """

        if self.camserver.open():
            self.connected_controller = 1
            self.camserver.set_timeout(5)
        else:
            self.connected_controller = 0

        return

    def disconnect(self):
        """
        Disconnets Azcam from the controller.
        """

        self.camserver.close()
        self.connected_controller = 0

        return

    def archon_command(self, Command):
        """
        Send a command to the Archon controller.
        """

        with self.lock:
            if not self.camserver.open():
                raise azcam.exceptions.AzcamError(
                    "Could not open connection to controller"
                )

            self.camserver.lastcmd_id = self.camserver.cmd_id
            self.camserver.cmd_id = (self.camserver.cmd_id + 1) & 0xFF
            preCmd = ">%02X" % (self.camserver.cmd_id)
            preResp = "<%02X" % (self.camserver.cmd_id)
            cmd = preCmd + Command
            if self.verbosity > 2:
                print("===>", cmd)

            self.camserver.send(cmd, "\r\n")

            if Command not in ["WARMBOOT", "REBOOT"]:
                reply = self.camserver.recv(-1)
                if self.verbosity > 2:
                    print("<===", reply[:40])
                status = reply.split(" ")[0]

                # check if the reply is synchronized
                if status[0:3] == preResp:
                    return reply[3:]
                else:
                    if reply[0] == "?":
                        raise azcam.exceptions.AzcamError("Archon response not valid")
                    else:
                        raise azcam.exceptions.AzcamError("Archon response out of sync")

        return None  # no Archon reponse is OK

    def archon_bin_command(self, command):
        """
        Send binary command to the Archon controller.
        """

        with self.lock:
            self.camserver.lastcmd_id = self.camserver.cmd_id
            self.camserver.cmd_id = (self.camserver.cmd_id + 1) & 0xFF

            preCmd = ">%02X" % (self.camserver.cmd_id)
            cmd = preCmd + command

            self.camserver.send(cmd, "\r\n")

        return

    def initialize(self):
        """
        Initializes the Archon controller.
        """

        if self.is_initialized:
            return

        # connect to controller
        self.connect()

        # load configuration file
        self.update_config_data(self.reset_flag)

        # set pars to stop exposure cycle and then load them
        self.set_continuous_exposures(0)
        self.set_exposures(0)
        self.set_int_ms(0)
        self.set_no_int_ms(0)
        self.load_params()
        self.resettiming()

        # check power status
        self.get_power_status()

        if self.power_status in ["OFF", "NOT_CONFIGURED"]:
            azcam.log("Power status: ", self.power_status, level=2)
            self.power_on(1)
        elif self.power_status != "ON":
            raise azcam.exceptions.AzcamError(f"Bad power status: {self.power_status}")

        self.is_initialized = 1

        return

    def reset(self):
        """
        Resets controller.
        """

        self.is_reset = 0

        # initialize controller
        self.initialize()

        self.is_reset = 1

        return

    def reset_controller(self):
        """
        Resets controller.
        Send REBOOT command + loads config file + sets Power ON.
        """

        # Connect to controller
        self.connect()
        if not self.connected_controller:
            raise azcam.exceptions.AzcamError("coud not connect to controller")

        self.warmboot()
        self.disconnect()

        # Wait before trying to send any command to the controller
        time.sleep(3)

        # Wait up to 30 sec for the controller to finish the boot sequence
        cnt = 0
        while cnt < 30:
            self.connect()
            if not self.connected_controller:
                time.sleep(1)
                continue

            try:
                self.get_status()
            except ConnectionResetError:
                time.sleep(1)
                pass
            if self.status_valid == 1:
                cnt = 30
            else:
                cnt += 1

        if self.status_valid != 1:
            raise azcam.exceptions.AzcamError("Controller reboot error")

        # reconnect to controller
        self.connect()

        # load configuration file
        self.update_config_data(1)

        # apply all config data
        self.apply_all()
        time.sleep(1)

        # set some pars
        self.set_continuous_exposures(0)

        # check power status
        reply = self.get_power_status()

        if reply in ["OFF", "NOT_CONFIGURED"]:
            self.power_on(1)
        else:
            raise azcam.exceptions.AzcamError("Bad controller power status")

        return

    def reboot(self):
        """
        Send REBOOT command.
        """

        cmd = "REBOOT"
        self.archon_command(cmd)

        return

    def warmboot(self):
        """
        Send WARMBOOT command.
        """

        cmd = "WARMBOOT"
        self.archon_command(cmd)

        return

    def get_power_status(self):
        """
        Get power status: ON, OFF, NOT_CONFIGURED, UNKNOWN, INTERMEDIATE, STANDBY.
        """

        self.get_status()
        self.power_status = self.power_values[int(self.status_power)]

        return self.power_status

    def get_status_valid(self):
        """
        Get status_valid value.
        Last change: 21Dec2016 Zareba
        """

        self.get_status()

        return self.status_valid

    def get_status(self):
        """
        Get status value.
        """

        cmd = "STATUS"
        reply = self.archon_command(cmd)

        # if reply == "OK":
        retVal = reply
        self.status = retVal.split(" ")
        self.dict_status = {}

        # Create a dictionary
        for item in self.status:
            if len(item) > 0:
                line = item.split("=")
                self.dict_status[line[0]] = line[1]

        # Update status keywords
        StatusDict = {}
        for line in self.status:
            if len(line) > 0:
                item = line.split("=")
                StatusDict[item[0]] = item[1]

        self.status_valid = int(StatusDict["VALID"])
        self.status_count = StatusDict["COUNT"]
        self.status_log = StatusDict["LOG"]
        self.status_power = StatusDict["POWER"]
        self.status_backplane_temp = StatusDict["BACKPLANE_TEMP"]

        return self.dict_status

    def get_frame(self):
        """
        Get and updates frame status value.
        """

        cmd = "FRAME"

        try:
            reply = self.archon_command(cmd)
        except socket.timeout:
            azcam.log("Socket timeout for FRAME command")
            return

        retVal = reply
        self.frame = retVal.split(" ")

        # Update frame status values

        self.dict_frame = {}

        # Create a dictionary
        for item in self.frame:
            if len(item) > 0:
                line = item.split("=")
                self.dict_frame[line[0]] = line[1]

        return self.dict_frame

    def get_size(self):
        """
        Get pixels and lines.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Config data not loaded")
        valPixels = 0
        valLines = 0

        indxParam = self.dict_wconfig[self.dict_params["Pixels"]]
        cmd = "RCONFIG%04X" % (indxParam & 0xFFFF)
        reply = self.archon_command(cmd)

        if len(reply) > 0:
            paramStr = reply.split("=")
            if len(paramStr) == 3:
                valPixels = int(paramStr[2])
            else:
                raise azcam.exceptions.AzcamError("Parameter error")
        else:
            raise azcam.exceptions.AzcamError("Parameter not found")

        indxParam = self.dict_wconfig[self.dict_params["Lines"]]
        cmd = "RCONFIG%04X" % (indxParam & 0xFFFF)
        reply = self.archon_command(cmd)

        if len(reply) > 0:
            paramStr = reply.split("=")
            if len(paramStr) == 3:
                valLines = int(paramStr[2])
            else:
                raise azcam.exceptions.AzcamError("Parameter error")
        else:
            raise azcam.exceptions.AzcamError("Parameter not found")

        return [valPixels, valLines]

    def set_size(self, Pixels, Lines):
        """
        Sets pixels and lines.
        """

        if len(self.dict_config) <= 0:
            raise azcam.exceptions.AzcamError("Config data not loaded")
        if self.lines != Lines:
            self.lines = Lines

            # update config dictionary
            self.dict_config[self.dict_params["Lines"]] = "Lines=%s" % (Lines)
            self.dict_config["LINECOUNT"] = Lines

            # update Archons Lines value
            indxParam = self.dict_wconfig[self.dict_params["Lines"]]
            cmd = "WCONFIG%04X%s=%s" % (
                indxParam & 0xFFFF,
                self.dict_params["Lines"],
                "Lines=" + str(Lines),
            )
            self.archon_command(cmd)

            indxParam = self.dict_wconfig["LINECOUNT"]
            cmd = "WCONFIG%04X%s=%s" % (indxParam & 0xFFFF, "LINECOUNT", str(Lines))
            self.archon_command(cmd)

        if self.pixels != Pixels:
            self.pixels = Pixels

            # update config dictionary
            self.dict_config[self.dict_params["Pixels"]] = "Pixels=%s" % (Pixels)
            self.dict_config["PIXELCOUNT"] = Pixels

            # update Archons Pixels value
            indxParam = self.dict_wconfig[self.dict_params["Pixels"]]
            cmd = "WCONFIG%04X%s=%s" % (
                indxParam & 0xFFFF,
                self.dict_params["Pixels"],
                "Pixels=" + str(Pixels),
            )
            self.archon_command(cmd)

            indxParam = self.dict_wconfig["PIXELCOUNT"]
            cmd = "WCONFIG%04X%s=%s" % (
                indxParam & 0xFFFF,
                "PIXELCOUNT",
                str(Pixels),
            )
            self.archon_command(cmd)

        return

    def load_params(self):
        """
        Send LOADPARAMS command.
        """

        cmd = "LOADPARAMS"
        self.archon_command(cmd)

        return

    def load_timing(self):
        """
        Send LOADTIMING command.
        """

        cmd = "LOADTIMING"
        self.archon_command(cmd)

        return

    def upload_config(self):
        """
        Uploads configuration data to the controller.
        """

        # start the counter
        cnt = 0x0000

        azcam.log("Uploading configuration data to controller", level=1)

        # clean the Write Config dictionary
        self.dict_wconfig = {}

        if len(self.config_data) == 0:
            raise azcam.exceptions.AzcamError("No configuration data")

        self.poll(0)

        # CLEARCONFIG
        cmd = "CLEARCONFIG"
        self.archon_command(cmd)

        # WCONFIG values
        for line in self.config_data:
            item = line.split("=")
            if len(item) >= 2:
                cmd = "WCONFIG%04X%s=%s" % (
                    cnt & 0xFFF,
                    item[0],
                    self.dict_config[item[0]].replace('"', ""),
                )
                self.dict_wconfig[item[0]] = cnt
                self.archon_command(cmd)
                cnt += 1

        self.poll(1)

        return

    def read_config_file(self, filename):
        """
        Read Archon configuration file and parse data into dictionaries.
        """

        self.config_data = []
        self.dict_wconfig = {}

        # Get configuration data from the timing file
        with open(self.timing_file, "r") as f:
            fBuff = f.read()
            sBuff = fBuff.split("\n")
            fLen = len(sBuff)

            # Extract CONFIG data
            indx = 0
            copy = 0
            pos = 0

            for line in sBuff:
                if line == "[CONFIG]":
                    copy = 1

                indx += 1
                if copy == 1 and indx < fLen and len(sBuff[indx]) > 0:
                    self.config_data.append(sBuff[indx].replace("\\", "/"))
                    self.dict_wconfig[self.config_data[pos].split("=")[0]] = pos
                    pos += 1

            self.config_lines_cnt = pos

        azcam.log(f"Read {self.config_lines_cnt} configuration lines", level=3)

        return

    def update_config_data(self, mode=0):
        """
        Sets configuration data:
        if mode = 0 get config data from the Archon controller, populate dictionaries.
        if mode = 1 read configuration file and then send config data to the Archon controller, populate dictionaries.
        """

        self.config_data = []
        self.dict_wconfig = {}
        self.dict_config = {}
        self.dict_params = {}
        self.dict_taplines = {}

        self.config_params = []

        self.config_lines_cnt = 0

        # get configuration data from file
        if mode == 1:
            self.read_config_file(self.timing_file)

        # get configuration data from the Archon controller
        else:
            azcam.log("Downloading configuration data from controller.", level=2)
            cnt = 0x0000
            endCfg = 0

            while endCfg != 1:
                cmd = "RCONFIG%04X" % (cnt & 0xFFFF)
                reply = self.archon_command(cmd)
                cnt += 1

                if len(reply) > 0:
                    self.config_data.append(reply)
                    self.dict_wconfig[reply.split("=")[0]] = cnt - 1
                else:
                    endCfg = 1

            self.config_lines_cnt = len(self.config_data)

        # Create a config directory with parameters:value pairs
        for item in self.config_data:
            if len(item) > 0:
                if item[0] != "[":
                    indx = item.find("=")
                    if indx > 0:
                        self.dict_config[item[:indx]] = item[indx + 1 :]

        # update parameters dictionaries
        paramCnt = int(self.dict_config["PARAMETERS"])
        for param in range(0, paramCnt):
            paramStr = "PARAMETER" + str(param)
            paramName = self.dict_config[paramStr].split("=")[0].replace('"', "")
            self.config_params.append(
                self.config_data[int(self.dict_wconfig[paramStr])]
            )
            self.dict_params[paramName] = paramStr

        # update configuration data
        firstParam = self.dict_wconfig["PARAMETER0"]
        for paramPos in range(0, paramCnt):
            self.config_data[firstParam + paramPos] = self.config_params[paramPos]
            item = self.config_params[paramPos]
            indx = item.find("=")
            if indx > 0:
                self.dict_config[item[:indx]] = item[indx + 1 :]

        # Update number of taplines - some lines might be empty
        cnt = 0
        tapLinesCnt = int(self.dict_config["TAPLINES"])
        for param in range(0, tapLinesCnt):
            tapLine = "TAPLINE" + str(param)
            tapLineVal = self.dict_config[tapLine].split("=")[0].replace('"', "")
            if len(tapLineVal) > 0:
                cnt += 1
                self.dict_taplines[tapLine] = tapLineVal
        self.tap_lines = cnt

        # extract exposure settings
        cont_exposures = (
            self.dict_config[self.dict_params["ContinuousExposures"]]
            .replace('"', "")
            .split("=")
        )
        self.cont_exposures = int(cont_exposures[1])

        exposures = (
            self.dict_config[self.dict_params["Exposures"]].replace('"', "").split("=")
        )
        self.exposures = int(exposures[1])

        sweep_cnt = (
            self.dict_config[self.dict_params["SweepCount"]].replace('"', "").split("=")
        )
        self.sweep_cnt = int(sweep_cnt[1])

        IntMS = self.dict_config[self.dict_params["IntMS"]].replace('"', "").split("=")
        self.int_ms = int(IntMS[1])

        NoIntMS = (
            self.dict_config[self.dict_params["NoIntMS"]].replace('"', "").split("=")
        )
        self.noint_ms = int(NoIntMS[1])

        # Config data is valid
        self.config_ok = 1

        if mode == 1:
            self.upload_config()

            # apply all config data
            self.apply_all()
            time.sleep(1)

            # set pars for exposures
            self.set_continuous_exposures(0)  # was 0

            # power on
            reply = self.get_power_status()
            if reply == "OFF" or reply == "NOT_CONFIGURED":
                self.power_on(1)
            else:
                raise azcam.exceptions.AzcamError(
                    "Power status not OFF or NOT_CONFIGURED"
                )

        return

    def power_on(self, wait=1):
        """
        Turns power on
        """

        # if already on, don't try due to status error
        reply = self.get_power_status()
        if reply == "ON":
            return

        cmd = "POWERON"

        self.archon_command(cmd)

        if wait == 1:
            # check power status - wait up to 10 second

            cnt = 0
            powerOK = 0
            while cnt < 10:
                self.get_power_status()
                if self.power_status == "ON":
                    # exit
                    cnt = 10
                    powerOK = 1
                else:
                    # wait
                    cnt += 1
                    time.sleep(1)

            if powerOK == 0:
                azcam.exceptions.warning("Controller poower is OFF")
            else:
                return
        else:
            return

    def power_off(self):
        """
        Turns power off.
        """

        cmd = "POWEROFF"

        self.archon_command(cmd)

        return

    def apply_all(self):
        """
        Send APPLYALL command to the Archon controller.
        """

        cmd = "APPLYALL"

        return self.archon_command(cmd)

    def update_cds(self, ucds=None):
        """
        Updates TAPLINES values based on space-delimited string.
        """

        if ucds is not None:
            self.ucds = ucds

        # temp = self.ucds[1 : len(self.ucds) - 1]

        nCDS = azcam.utils.parse(self.ucds)
        print("nCDS", nCDS)

        self.cds = []
        for item in nCDS:
            if len(item) > 5:
                self.cds.append(item)

        self.set_cds()

        return

    def set_cds(self):
        """
        Sets TAPLINES values.
        """

        lenCDS = len(self.cds)

        for tapLinesCnt in range(lenCDS):
            tapLine = "TAPLINE" + str(tapLinesCnt)
            indx = self.dict_wconfig[tapLine]
            cmd = "WCONFIG%04X%s=%s" % (indx & 0xFFFF, tapLine, self.cds[tapLinesCnt])
            self.archon_command(cmd)

        self.apply_cds()

        return

    def get_cds(self):
        """
        Get TAPLINES values and stores them in the self.rcds array.
        """

        self.rcds = []

        for tapLines in range(self.tap_lines):
            indxParam = self.dict_wconfig["TAPLINE" + str(tapLines)]
            cmd = "RCONFIG%04X" % (indxParam & 0xFFFF)

            reply = self.archon_command(cmd)
            respLine = reply.split("=")
            self.rcds.append(respLine[1])

        return self.rcds

    def apply_cds(self):
        """
        Send APPLYCDS command to the controller.
        CSS/Deint values are updated.
        """

        cmd = "APPLYCDS"

        return self.archon_command(cmd)

    def get_continuous_exposures(self):
        """
        Get Continuous Exposure value.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        indxParam = self.dict_wconfig[self.dict_params["ContinuousExposures"]]
        cmd = "RCONFIG%04X" % (indxParam & 0xFFFF)
        reply = self.archon_command(cmd)

        if len(reply) > 0:
            paramStr = reply.split("=")
            if len(paramStr) == 3:
                return int(paramStr[2])
            else:
                raise azcam.exceptions.AzcamError("Parameter error")
        else:
            raise azcam.exceptions.AzcamError("Parameter not found")

        return

    def set_continuous_exposures(self, cont_exp):
        """
        Sets Continuous Exposure value.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        self.cont_exposures = int(cont_exp)

        # update config dictionary
        self.dict_config[self.dict_params["ContinuousExposures"]] = (
            "ContinuousExposures=%s" % (self.cont_exposures)
        )

        # update Archons CountinuousExposures value
        indxParam = self.dict_wconfig[self.dict_params["ContinuousExposures"]]
        cmd = "WCONFIG%04X%s=%s" % (
            indxParam & 0xFFFF,
            self.dict_params["ContinuousExposures"],
            "ContinuousExposures=" + str(self.cont_exposures),
        )
        self.archon_command(cmd)

        return

    def get_frame_number(self):
        """
        Get Frame number.
        """

        f1 = self.dict_frame["BUF1FRAME"]
        f2 = self.dict_frame["BUF2FRAME"]
        f3 = self.dict_frame["BUF3FRAME"]

        if f2 > f1 and f2 > f3:
            max = f2
        elif f2 > f1 or f1 <= f3:
            max = f3
        else:
            max = f1
        return max

    def get_parameters(self):
        """
        Get parameters.
        """

        self.parameters = {}
        if len(self.dict_config) > 0:
            cnt = int(self.dict_config["PARAMETERS"])

            for indx in range(cnt):
                param = "PARAMETER" + str(indx)
                parname = self.dict_config[param].split("=")[0]
                parvalue = self.dict_config[param].split("=")[1]
                self.parameters[parname] = parvalue

            return self.parameters
        else:
            raise azcam.exceptions.AzcamError("Configuration data error")

    def get_parameter(self, Param):
        """
        Get a parameter by name.
        Returns None if not found.
        """

        self.get_parameters()

        return self.parameters.get(Param)

    def set_parameter(self, Param, value):
        """
        Sets parameter.
        """

        found = 0
        if len(self.dict_config) > 0:
            cnt = int(self.dict_config["PARAMETERS"])

            for indx in range(cnt):
                param = "PARAMETER" + str(indx)
                paramName = self.dict_config[param].split("=")[0]  # new
                if paramName == Param:
                    self.dict_config[param] = Param + "=" + str(value)
                    found = 1
                    break

            if found == 1:
                return
            else:
                raise azcam.exceptions.AzcamError("Parameter not found")
        else:
            raise azcam.exceptions.AzcamError("Configuration error")

    def get_exposures(self):
        """
        Get number of exposures.
        """

        if self.config_ok:
            indxParam = self.dict_wconfig[self.dict_params["Exposures"]]
            cmd = "RCONFIG%04X" % (indxParam & 0xFFFF)
            reply = self.archon_command(cmd)

            if len(reply) > 0:
                paramStr = reply.split("=")
                if len(paramStr) == 3:
                    return int(paramStr[2])
                else:
                    raise azcam.exceptions.AzcamError("Parameter error")
            else:
                raise azcam.exceptions.AzcamError("Parameter not found")
        else:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        return self.exposures

    def set_exposures(self, Exp):
        """
        Sets number of exposures.
        """

        if self.exposures != Exp:
            if not self.config_ok:
                raise azcam.exceptions.AzcamError("Configuration data not loaded")

            self.exposures = Exp

            # update config dictionary
            self.dict_config[self.dict_params["Exposures"]] = "Exposures=%s" % (
                self.exposures
            )

            # update Archons Exposures value
            indxParam = self.dict_wconfig[self.dict_params["Exposures"]]
            cmd = "WCONFIG%04X%s=%s" % (
                indxParam & 0xFFFF,
                self.dict_params["Exposures"],
                "Exposures=" + str(Exp),
            )
            self.archon_command(cmd)

        return

    def get_archon_status(self):
        """
        Get Archon controller status: 0 = UNKNOWN, 1 = IDLE, 2 = EXPOSE, 3 = READY, 4 = FETCH, 5 = DONE.
        """

        self.get_frame()

        return self.archon_status

    def poll(self, mode):
        """
        Send POLLON or POLLOFF command.
        """

        cmd = "POLLON" if mode else "POLLOFF"
        return self.archon_command(cmd)

    def bias_poll(self, mode):
        """
        Send BIASPOLLON or BIASPOLLOFF command.
        """

        cmd = "BIASPOLLON" if mode else "BIASPOLLOFF"
        return self.archon_command(cmd)

    def get_exposuretime(self):
        """
        Get exposure time from the controller.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        indxParam = self.dict_wconfig[self.dict_params["IntMS"]]
        cmd = "RCONFIG%04X" % (indxParam & 0xFFFF)
        reply = self.archon_command(cmd)

        if len(reply) > 0:
            paramStr = reply.split("=")
            if len(paramStr) == 3:
                return int(paramStr[2]) / 1000.0
            else:
                raise azcam.exceptions.AzcamError("IntMS parameter error")
        else:
            raise azcam.exceptions.AzcamError("IntMS parameter not found")

        return

    def set_exposuretime(self, ExpTimeMS):
        """
        Set exposure time variable (millisecs).
        This is used for exposure count down.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        self.exp_time_ms = int(ExpTimeMS)
        self.int_ms = int(ExpTimeMS)

        # update config dictionary
        self.dict_config[self.dict_params["IntMS"]] = "IntMS=%s" % (self.int_ms)

        # update Archons IntMS value
        indxParam = self.dict_wconfig[self.dict_params["IntMS"]]
        cmd = "WCONFIG%04X%s=%s" % (
            indxParam & 0xFFFF,
            self.dict_params["IntMS"],
            "IntMS=" + str(ExpTimeMS),
        )
        self.archon_command(cmd)

        return

    def get_int_ms(self):
        """
        Get IntMS value. The value should be the same as the Archons IntMS.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        indxParam = self.dict_wconfig[self.dict_params["IntMS"]]
        cmd = "RCONFIG%04X" % (indxParam & 0xFFFF)
        reply = self.archon_command(cmd)

        if len(reply) > 0:
            paramStr = reply.split("=")
            if len(paramStr) == 3:
                return int(paramStr[2])
            else:
                raise azcam.exceptions.AzcamError("Parameter error")
        else:
            raise azcam.exceptions.AzcamError("Parameter not found")

        return

    def set_int_ms(self, IntMS):
        """
        Set IntMS value.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        self.exp_time_ms = int(IntMS)
        self.int_ms = int(IntMS)

        # special for long exposure times - IN PROGRESS
        ms = IntMS
        mul = 1
        IntMul = mul
        tmax = (1 << 20) - 2  # Archon parameter word length tmax=1048574
        if ms > tmax:
            mul = (ms - 1) / tmax + 1  # makes mul > 0
            ms = ms / mul
            IntMS = int(ms)
            IntMul = int(mul)

        # update config dictionary
        self.dict_config[self.dict_params["IntMS"]] = "IntMS=%s" % (IntMS)
        self.dict_config[self.dict_params["IntMul"]] = "IntMul=%s" % (IntMul)

        # update Archons IntMS value
        indxParam = self.dict_wconfig[self.dict_params["IntMS"]]
        cmd = "WCONFIG%04X%s=%s" % (
            indxParam & 0xFFFF,
            self.dict_params["IntMS"],
            "IntMS=" + str(IntMS),
        )
        self.archon_command(cmd)

        # update Archons IntMul value
        indxParam = self.dict_wconfig[self.dict_params["IntMul"]]
        cmd = "WCONFIG%04X%s=%s" % (
            indxParam & 0xFFFF,
            self.dict_params["IntMul"],
            "IntMul=" + str(IntMul),
        )
        self.archon_command(cmd)

        return

    def set_pocket_pumping(self, flag):
        """
        Set Parallel pocket pumping flag.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        # update config dictionary
        self.dict_config[self.dict_params["ParallelPumping"]] = "ParallelPumping=%s" % (
            str(flag)
        )

        # update Archons value
        indxParam = self.dict_wconfig[self.dict_params["ParallelPumping"]]
        cmd = "WCONFIG%04X%s=%s" % (
            indxParam & 0xFFFF,
            self.dict_params["ParallelPumping"],
            "ParallelPumping=" + str(flag),
        )
        self.archon_command(cmd)

        return

    def get_no_int_ms(self):
        """
        Get NoIntMS value. The value should be the same as the Archons NoIntMS.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        indxParam = self.dict_wconfig[self.dict_params["NoIntMS"]]
        cmd = "RCONFIG%04X" % (indxParam & 0xFFFF)
        reply = self.archon_command(cmd)

        if len(reply) > 0:
            paramStr = reply.split("=")
            if len(paramStr) == 3:
                return int(paramStr[2])
            else:
                raise azcam.exceptions.AzcamError("Parameter error")
        else:
            raise azcam.exceptions.AzcamError("Paramter not found")

        return

    def set_no_int_ms(self, NoIntMS):
        """
        Set NoIntMS value.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        self.noint_ms = int(NoIntMS)

        # special for long exposure times - IN PROGRESS
        ms = NoIntMS
        mul = 1
        NoIntMul = mul
        tmax = (1 << 20) - 2  # Archon parameter word length tmax=1048574
        if ms > tmax:
            mul = (ms - 1) / tmax + 1  # makes mul > 0
            ms = ms / mul
            NoIntMS = int(ms)
            NoIntMul = int(mul)

        # update config dictionary
        self.dict_config[self.dict_params["NoIntMS"]] = "NoIntMS=%s" % (NoIntMS)
        self.dict_config[self.dict_params["NoIntMul"]] = "NoIntMul=%s" % (NoIntMul)

        # update Archons NoIntMS value
        indxParam = self.dict_wconfig[self.dict_params["NoIntMS"]]
        cmd = "WCONFIG%04X%s=%s" % (
            indxParam & 0xFFFF,
            self.dict_params["NoIntMS"],
            "NoIntMS=" + str(NoIntMS),
        )
        self.archon_command(cmd)

        indxParam = self.dict_wconfig[self.dict_params["NoIntMul"]]
        cmd = "WCONFIG%04X%s=%s" % (
            indxParam & 0xFFFF,
            self.dict_params["NoIntMul"],
            "NoIntMul=" + str(NoIntMul),
        )
        self.archon_command(cmd)

        return

    def fastloadparam(self, parname, value):
        """
        Send FASTLOADPARAM command to controller.
        """

        self.archon_command(f"FASTLOADPARAM {parname} {value}")

        return

    def get_raw_enable(self):
        """
        Get RAWENABLE value.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        indxParam = self.dict_wconfig["RAWENABLE"]
        cmd = "RCONFIG%04X" % (indxParam & 0xFFFF)
        reply = self.archon_command(cmd)

        if len(reply) > 0:
            paramStr = reply.split("=")
            if len(paramStr) == 2:
                self.rawdata_enable = int(paramStr[1])
                return int(paramStr[1])
            else:
                raise azcam.exceptions.AzcamError("Parameter error")
        else:
            raise azcam.exceptions.AzcamError("Parameter not found")

    def set_raw_enable(self, RawEnable):
        """
        Set RAWENABLE value.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        self.rawdata_enable = RawEnable

        # update config dictionary
        self.dict_config["RAWENABLE"] = "%s" % (self.rawdata_enable)

        # update Archons RAWENABLE value
        indxParam = self.dict_wconfig["RAWENABLE"]
        cmd = "WCONFIG%04X%s=%s" % (
            indxParam & 0xFFFF,
            "RAWENABLE",
            str(self.rawdata_enable),
        )
        self.archon_command(cmd)
        self.apply_cds()

        return

    def get_raw_channel(self):
        """
        Get RAWSEL value (raw channel selection).
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        indxParam = self.dict_wconfig["RAWSEL"]
        cmd = "RCONFIG%04X" % (indxParam & 0xFFFF)
        reply = self.archon_command(cmd)

        if len(reply) > 0:
            paramStr = reply.split("=")
            if len(paramStr) == 2:
                return int(paramStr[1]) + 1
            else:
                raise azcam.exceptions.AzcamError("Parameter error")
        else:
            raise azcam.exceptions.AzcamError("Parameter not found")

        return

    def set_raw_channel(self, RawChannel):
        """
        Sets RAWSEL value  (raw channel selection).
        RAWSEL starts from 0.
        """

        if not self.config_ok:
            raise azcam.exceptions.AzcamError("Configuration data not loaded")

        self.rawdata_channel = RawChannel

        # update config dictionary
        self.dict_config["RAWSEL"] = "%s" % (self.rawdata_channel - 1)

        # update Archons RAWENABLE value
        indxParam = self.dict_wconfig["RAWSEL"]
        cmd = "WCONFIG%04X%s=%s" % (
            indxParam & 0xFFFF,
            "RAWSEL",
            str(self.rawdata_channel - 1),
        )
        self.archon_command(cmd)
        self.apply_cds()

        return

    def get_pixels_remaining(self):
        """
        Return number of remaining pixels to be read (counts down).
        """

        if self.dict_frame == {} or self.read_buffer == 0:
            return 0

        fp = azcam.db.tools["exposure"].get_focalplane()
        # numseramps = fp[0] * fp[2]
        # numparamps = fp[1] * fp[3]
        numseramps = fp[2]
        numparamps = fp[3]
        naxis1 = int(self.dict_frame[f"BUF{self.read_buffer}WIDTH"])
        naxis2 = int(self.dict_frame[f"BUF{self.read_buffer}HEIGHT"])
        total_pixels = naxis1 * naxis2

        pixels = int(self.dict_frame[f"BUF{self.read_buffer}PIXELS"])
        lines = int(self.dict_frame[f"BUF{self.read_buffer}LINES"])

        # this is for split pit mode
        if int(self.dict_config["FRAMEMODE"]) == 2:
            lines = lines * 2
        # pixels_read = lines * naxis1 + numparamps * pixels
        pixels_read = lines * naxis1 + (numparamps + numseramps) * pixels
        pixels_remaining = max(0, int(total_pixels - pixels_read))

        return pixels_remaining

    def update_exposuretime_remaining(self):
        """
        Returns exposure time remaining in seconds.
        Last change: 13Jan2017 Zareba
        """

        dt = int(self.dict_frame["TIMER"], 16) - self.exp_timer
        et = self.int_ms / 1000.0 + self.noint_ms / 1000.0 - dt / 100000000.0

        if et < 0:
            et = 0

        return et

    def start_exposure(self, wait=1):
        """
        Start exposure.
        """

        # Set exposure state to UNKNOWN
        self.archon_status = self.EXP_UNKNOWN

        # Check frame status
        self.get_frame()

        # Save current frame numbers
        self.currframe1 = self.dict_frame["BUF1FRAME"]
        self.currframe2 = self.dict_frame["BUF2FRAME"]
        self.currframe3 = self.dict_frame["BUF3FRAME"]

        # Start exposure -> send LOADPARAMS command for single exposure mode
        self.set_exposures(0)
        self.load_params()  # load exposure times and other pars
        self.fastloadparam("EXPOSURES", 1)
        self.frame_time = time.time()

        # Exposure start time -> used to determine time out
        self.exp_start = time.time()

        # Exposure start time (internal Archon controller timer - resolution 10 ns)
        self.get_frame()
        self.exp_timer = int(self.dict_frame["TIMER"], 16)

        int_time = (int(self.int_ms) + int(self.noint_ms)) / 1000

        # if no wait, just set status and exit
        if wait == 0:
            self.archon_status = self.EXP_EXPOSE
            return

        # frame number changes when integration is over
        stop = 0
        self.newframe = 0

        # Set exposure flag to INTEGRATING
        azcam.db.tools["exposure"].exposure_flag = azcam.db.tools[
            "exposure"
        ].exposureflags["EXPOSING"]

        # wait for frame to change in buffers
        if int_time > 0:
            azcam.log("Integrating", level=1)
        while stop == 0:
            # Get frame and update frame dictionary
            self.get_frame()

            # Check in a new frame is available
            if self.currframe1 != self.dict_frame["BUF1FRAME"]:
                self.newframe = 1
                stop = 1

            if self.currframe2 != self.dict_frame["BUF2FRAME"]:
                self.newframe = 2
                stop = 1

            if self.currframe3 != self.dict_frame["BUF3FRAME"]:
                self.newframe = 3
                stop = 1

            if int_time > 0:
                pass
                # azcam.log(f"Integrating: {(time.time() - self.exp_start):.1f} secs", level=2)

            # check for abort
            if (
                azcam.db.tools["exposure"].exposure_flag
                == azcam.db.tools["exposure"].exposureflags["ABORT"]
            ):
                stop = 1

            # Check if time out occured
            if not stop:
                if 0:  # 0 is no timeout
                    if time.time() > (self.frame_time + int_time + 90):  # long for 10k
                        self.newframe = -1
                        self.read_buffer = -1
                        azcam.exceptions.warning("Timed out waiting for integration")
                        stop = 1
                time.sleep(0.5)

        # check for abort
        if (
            azcam.db.tools["exposure"].exposure_flag
            == azcam.db.tools["exposure"].exposureflags["ABORT"]
        ):
            self.archon_status = self.EXP_DONE
            azcam.exceptions.warning("Exposure aborted")
            return

        # Set exposure flag to READOUT
        azcam.db.tools["exposure"].exposure_flag = azcam.db.tools[
            "exposure"
        ].exposureflags["READOUT"]

        self.read_buffer = self.newframe

        # Wait for readout to complete
        azcam.log("Reading", level=1)
        if self.newframe == 0:
            self.archon_status = self.EXP_UNKNOWN
            raise azcam.exceptions.AzcamError("New frame is not ready")

        self.read_time = time.time()
        frameStatus = "BUF%dCOMPLETE" % (self.read_buffer)
        dataReady = 0
        cnt = 0

        while dataReady == 0 and cnt < 500:
            # Get frame and update frame dictionary
            time.sleep(0.50)
            self.get_frame()

            # Check if frame is ready
            ready = self.dict_frame[frameStatus]
            if int(ready) == 1:
                dataReady = 1
            else:
                cnt += 1

            if 1:
                azcam.log(
                    f"Reading: {(time.time() - self.read_time):.1f} secs", level=2
                )

            # check for abort
            if (
                azcam.db.tools["exposure"].exposure_flag
                == azcam.db.tools["exposure"].exposureflags["ABORT"]
            ):
                dataReady = -1
                break

        # record actual exposure time - archon time stamp valid for INT only
        et = 0
        if self.int_ms > 0:
            t1 = int(self.dict_frame[f"BUF{self.read_buffer}RETIMESTAMP"], 16)
            t2 = int(self.dict_frame[f"BUF{self.read_buffer}FETIMESTAMP"], 16)
            et = (t2 - t1) / 1.0e8
        elif self.noint_ms > 0:
            et = self.noint_ms / 1000.0
        azcam.db.tools["exposure"].exposure_time_actual = et

        if dataReady == 1:
            self.archon_status = self.EXP_DONE
            return
        elif dataReady == -1:
            self.archon_status = self.EXP_DONE
            azcam.exceptions.warning("Exposure aborted")
        else:
            raise azcam.exceptions.AzcamError("New frame data is not ready")

        return

    def download_config(self):
        """
        Downloads config data from the Archon controller.
        """

        cnt = 0x0000

        self.ConfigArchon = []

        endCfg = 0

        while endCfg != 1:
            cmd = "RCONFIG%04X" % (cnt & 0xFFFF)
            reply = self.archon_command(cmd)
            cnt += 1
            if len(reply) > 0:
                self.ConfigArchon.append(reply[1])
            else:
                endCfg = 1

        self.ConfigArchonCnt = len(self.ConfigArchon)

        return

    def resettiming(self):
        """
        Reset controller timing cores.
        """

        self.archon_command("RESETTIMING")

        return

    # ROI

    def set_roi(self):
        """
        Sets the ROI parameters values in the controller based on focalplane parameters.
        Sends parameters to the controller.
        """

        try:
            self.write_controller_roi()
        except azcam.exceptions.AzcamError as e:
            pass  # can't update until controller is reset

        return

    def write_controller_roi(self):
        """
        Set controller ROI values.
        """

        # self.detpars.first_col = self.image.focalplane.first_col
        # self.detpars.last_col = self.image.focalplane.last_col
        # self.detpars.first_row = self.image.focalplane.first_row
        # self.detpars.last_row = self.image.focalplane.last_row
        # self.detpars.col_bin = self.image.focalplane.col_bin
        # self.detpars.row_bin = self.image.focalplane.row_bin

        # self.detpars.xunderscan = self.image.focalplane.xunderscan
        # self.detpars.xskip = self.image.focalplane.xskip
        # self.detpars.xpreskip = self.image.focalplane.xpreskip
        # self.detpars.xdata = self.image.focalplane.xdata
        # self.detpars.xpostskip = self.image.focalplane.xpostskip
        # self.detpars.xoverscan = self.image.focalplane.xoverscan
        # self.detpars.yunderscan = self.image.focalplane.yunderscan
        # self.detpars.yskip = self.image.focalplane.yskip
        # self.detpars.ypreskip = self.image.focalplane.ypreskip
        # self.detpars.ydata = self.image.focalplane.ydata
        # self.detpars.ypostskip = self.image.focalplane.ypostskip
        # self.detpars.yoverscan = self.image.focalplane.yoverscan

        # self.detpars.numcols_amp = self.image.focalplane.numcols_amp
        # self.detpars.numcols_overscan = self.image.focalplane.numcols_overscan
        # self.detpars.numviscols_amp = self.image.focalplane.numviscols_amp
        # self.detpars.numviscols_image = self.image.focalplane.numviscols_image
        # self.detpars.numrows_amp = self.image.focalplane.numrows_amp
        # self.detpars.numrows_overscan = self.image.focalplane.numrows_overscan
        # self.detpars.numvisrows_amp = self.image.focalplane.numvisrows_amp
        # self.detpars.numvisrows_image = self.image.focalplane.numvisrows_image
        # self.detpars.numpix_amp = self.image.focalplane.numpix_amp
        # self.detpars.numcols_det = self.image.focalplane.numcols_det
        # self.detpars.numrows_det = self.image.focalplane.numrows_det
        # self.detpars.numpix_det = self.image.focalplane.numpix_det
        # self.detpars.numpix_image = self.image.focalplane.numpix_image
        # self.detpars.numcols_image = self.image.focalplane.numcols_image
        # self.detpars.numrows_image = self.image.focalplane.numrows_image
        # self.detpars.numbytes_image = self.image.focalplane.numbytes_image
        # self.detpars.xflush = self.image.focalplane.xflush
        # self.detpars.yflush = self.image.focalplane.yflush

        xpreskip = self.detpars.xpreskip + self.detpars.xskip
        ypreskip = self.detpars.ypreskip + self.detpars.yskip

        # new 4/25

        roi_pars = {
            "PreSkipPixels": xpreskip,
            "HorizontalBinning": self.detpars.col_bin - 1,
            "PostSkipPixels": self.detpars.xpostskip,
            "OverScanPixels": self.detpars.xoverscan,
            "PreSkipLines": ypreskip,
            "VerticalBinning": self.detpars.row_bin,
            "PostSkipLines": self.detpars.ypostskip,
            "OverScanLines": self.detpars.yoverscan,
        }

        if not self.config_ok:
            # raise azcam.exceptions.AzcamError("Configuration data not loaded")
            return

        # update config dictionary
        for par in roi_pars:
            self.dict_config[self.dict_params[par]] = f"{par}={roi_pars[par]}"

            # update Archon value
            indxParam = self.dict_wconfig[self.dict_params[par]]
            cmd = "WCONFIG%04X%s=%s" % (
                indxParam & 0xFFFF,
                self.dict_params[par],
                f"{par}={roi_pars[par]}",
            )
            self.archon_command(cmd)

        # also update pixelcount and linecount (per tap)
        pixelcount = (
            self.detpars.xunderscan + self.detpars.xdata + self.detpars.xoverscan
        )
        linecount = (
            self.detpars.yunderscan + self.detpars.ydata + self.detpars.yoverscan
        )
        # indxParam = self.dict_wconfig["LINECOUNT"]
        # cmd = "WCONFIG%04X%s=%s" % (indxParam & 0xFFFF, "LINECOUNT", str(linecount))
        # self.archon_command(cmd)

        # indxParam = self.dict_wconfig["PIXELCOUNT"]
        # cmd = "WCONFIG%04X%s=%s" % (
        #     indxParam & 0xFFFF,
        #     "PIXELCOUNT",
        #     str(pixelcount),
        # )
        # self.archon_command(cmd)

        self.set_size(pixelcount, linecount)
        self.apply_cds()
        self.load_params()

        return



"""
test instrument hardware - server-side
"""

import sys
import time

import azcam


def test_instrument(loops=2):
    """
    Test instrument hardware.
    """

    loops = int(loops)

    print("Testing instrument...")

    for i in range(loops):
        print("Reading header")
        reply = azcam.db.tools["instrument"].read_header()
        for x in reply:
            print(f"--> {x}")

        print("Reading wavelength")
        reply = azcam.db.tools["instrument"].get_wavelength()
        print(f"--> {reply:.3f}")

        print("Cycling shutter")
        reply = azcam.db.tools["instrument"].set_shutter(1)
        print("--> Opened")
        time.sleep(1)
        reply = azcam.db.tools["instrument"].set_shutter(0)
        print("--> Closed")

    print("Finished testing")

    return


if __name__ == "__main__":
    args = sys.argv[1:]
    test_instrument(*args)



"""
Contains the TempConASCOM class.
"""

import azcam
import azcam.exceptions
from azcam.tools.tempcon import TempCon


class TempConASCOM(TempCon):
    """
    Temperature control class for cameras using ASCOM.
    """

    def __init__(self, tool_id="tempcon", description=None):
        super().__init__(tool_id, description)

        self.temperature_ids = [0, 1]  # camtemp, dewtemp

        self.init_commands = []

    def define_keywords(self):
        """
        Defines and resets tempcon keywords.
        """

        self.set_keyword("CAMTEMP", 0.0, "Camera temperature", "float")
        self.set_keyword("DEWTEMP", 0.0, "Dewar temperature", "float")
        self.set_keyword("TEMPUNIT", "C", "Temperature units", "str")

        return

    def initialize(self):
        """
        Initialize CryoConM24TemperatureControl temperature controller.
        """

        if self.is_initialized:
            return
        if not self.is_enabled:
            azcam.exceptions.warning(f"{self.description} is not enabled")
            return

        # set control temp
        self.set_control_temperature(self.control_temperature)

        self.is_initialized = 1

        return

    def set_control_temperature(
        self, temperature: float = None, temperature_id: int = -1
    ):
        """
        Set control temperature in Celsius.
        Args:
            temperature: temperature to set
            temperature_id: temperature ID number
                * 0 is TempA
                * 1 is TempB
        """

        if temperature is None:
            temperature = self.control_temperature
        else:
            self.control_temperature = float(temperature)

        # turn cooler ON
        azcam.db.tools["tempcon"].set_cooler(1)

        azcam.db.tools["controller"].camera.SetCCDTemperature = temperature

        return

    def get_temperature(self, temperature_id: int = 0):
        """
        Reads temperatures from the tempcon controller.
        Args:
            temperature_id: temperature ID number
        Returns:
            temperature: temperature read
        """

        if not self.is_enabled:
            # azcam.exceptions.warning("Tempcon not enabled")
            return -999.9

        if not self.is_initialized:
            # azcam.exceptions.warning("Tempcon not initialized")
            return -999.9

        temperature_id = int(temperature_id)
        if temperature_id == 0:
            pass
        elif temperature_id == 1:
            pass
        else:
            raise azcam.exceptions.AzcamError("bad temperature_id in get_temperature")

        reply = azcam.db.tools["controller"].camera.CCDTemperature
        try:
            temp = float(reply)
        except ValueError:
            temp = self.bad_temp_value

        temp = self.apply_corrections(temp, temperature_id)

        return temp

    def get_cooler_power(self):
        """
        Reads cooler power.
        Returns:
            temperature: temperature read
        """

        return azcam.db.tools["controller"].camera.CoolerPower

    def set_cooler(self, state):
        """
        Sets cooler ON (1) of OFF (0).
        Args:
            state: 1 for ON, 0 for OFF
        """

        azcam.db.tools["controller"].camera.CoolerOn = state

        return



from dash import html, dcc, callback, Input, Output, State
import dash_bootstrap_components as dbc

import azcam


def filename_card():
    """
    Create filename card.
    """

    """
    Directory with browse
    Rootname
    Seq Num spinner
    Image Format NO
    
    Auto inc
    Overwrite
    Inc. seq
    Autoname
    test image
    """

    curfolder = "/data"

    folderselect = html.Div(
        [
            dbc.Label("Image folder"),
            dbc.Input(
                id="folderselect",
                placeholder="Enter image file folder...",
                type="text",
                style={"display": "inline-block"},
            ),
        ]
    )

    rootselect = html.Div(
        [
            dbc.Label("Image root name"),
            dbc.Input(
                id="rootselect",
                placeholder="Enter image root name...",
                type="text",
                style={"display": "inline-block"},
            ),
        ]
    )

    seqnumselect = html.Div(
        [
            dbc.Label("Image sequence number"),
            dbc.Input(
                id="seqnumselect",
                placeholder="Enter image sequence number...",
                type="number",
                min=0,
                style={"display": "inline-block"},
            ),
        ]
    )

    filenamechecks_input = dcc.Checklist(
        [
            {
                "label": "Auto increment",
                "value": "autoinc",
                "title": "check to auto increment image sequence number",
            },
            {
                "label": "Overwrite existing image",
                "value": "overwrite",
            },
            {
                "label": "Include sequence number",
                "value": "includeseqnum",
            },
            {
                "label": "Auto name",
                "value": "autoname",
                "title": "check to inlcude Object Type in image filename",
            },
        ],
        id="filename_checks",
        value=["overwrite"],
        inline=False,
    )
    checks_filename = html.Div(filenamechecks_input)

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("filename_checks", "value"),
        prevent_initial_call=True,
    )
    def image_test_callback(value):
        azcam.db.expserver.set_message(value)
        return value

    # apply button - all options on this page
    apply_filename_btn = dbc.Button(
        "Apply All",
        id="apply_filename_btn",
        style={"margin-top": "1em"},
        n_clicks=0,
    )

    @callback(
        Output("fileselect_out", "children", allow_duplicate=True),
        Input("apply_filename_btn", "n_clicks"),
        State("seqnumselect", "value"),
        State("rootselect", "value"),
        State("folderselect", "value"),
        prevent_initial_call=True,
    )
    def on_button_click_apply_filename(n, seqnum, root, folder):
        try:
            azcam.db.parameters.set_par("imagesequencenumber", seqnum)
            azcam.db.parameters.set_par("imageroot", root)
            azcam.db.parameters.set_par("imagefolder", folder)
        except Exception as e:
            print(e)
        return

    filename_card = dbc.Card(
        [
            dbc.CardHeader("Filename Parameters", style={"font-weight": "bold"}),
            dbc.CardBody(
                [
                    dbc.Row(
                        [
                            dbc.Col(
                                [
                                    folderselect,
                                    rootselect,
                                    seqnumselect,
                                ]
                            ),
                            dbc.Col(
                                [
                                    checks_filename,
                                ]
                            ),
                        ]
                    ),
                    dbc.Row(
                        [
                            apply_filename_btn,
                            html.Div(id="fileselect_out"),
                        ]
                    ),
                ]
            ),
        ]
    )

    return filename_card



"""
Contains the azcam database class for azcamserver.

There is only one instance of this class which is referenced as `azcam.db` and contains
temporary data for this current process.
"""

from azcam.database import AzcamDatabase
from azcam.parameters_server import ParametersServer
from azcam.cmdserver import CommandServer
from azcam.web.fastapi_server import WebServer
from azcam.header import System


class AzcamDatabaseServer(AzcamDatabase):
    """
    The azcam database class.
    """

    parameters: ParametersServer
    """parameters object"""

    cmdserver: CommandServer
    """command server object"""

    webserver: WebServer
    """webserver object"""

    servermode: str
    """server mode"""

    monitor = object
    """azcammonitor object"""

    default_tool: str = "api"
    """default tool for remote commands"""

    system: System
    """system header object"""



"""
Contains the TempConArchon class.
"""

import azcam
import azcam.exceptions
from azcam.tools.tempcon import TempCon


class TempConArchon(TempCon):
    """
    Defines the Archon temperature control tool.
    """

    def __init__(self, tool_id="tempcon", description=None):
        super().__init__(tool_id, description)

        self.num_temp_reads = 1
        self.heaterx_board = "MOD2"

        self.temperature_ids = [0, 1]  # camtemp, dewtemp

        return

    def define_keywords(self):
        """
        Defines and resets tempcon keywords.
        """

        self.set_keyword("CAMTEMP", 0.0, "Camera temperature", "float")
        self.set_keyword("DEWTEMP", 0.0, "Dewar temperature", "float")
        self.set_keyword("TEMPUNIT", "C", "Temperature units", "str")

        return

    def get_temperature(self, temperature_id=0):
        """
        Read a camera temperature.
        TemperaureID's are:
        0 => TEMPA
        1 => TEMPB
        2 => TEMPC
        """

        if not self.is_enabled:
            return -999.9

        if not self.is_initialized:
            return -999.9

        if not azcam.db.tools["controller"].heater_board_installed:
            return self.bad_temp_value

        if not azcam.db.tools["controller"].is_reset:
            return self.bad_temp_value

        temperature_id = int(temperature_id)
        if temperature_id == 0:
            Address = f"{self.heaterx_board}/TEMPA"
        elif temperature_id == 1:
            Address = f"{self.heaterx_board}/TEMPB"
        elif temperature_id == 2:
            Address = f"{self.heaterx_board}/TEMPC"
        else:
            raise azcam.exceptions.AzcamError("bad temperature_id in get_temperature")

        # Don't read hardware while exposure is in progess
        flag = azcam.db.tools["exposure"].exposure_flag
        if flag != azcam.db.tools["exposure"].exposureflags["NONE"]:
            return self.last_temps[temperature_id]

        # read temperature
        avetemp = 0
        for _ in range(self.num_temp_reads):
            temp = float(azcam.db.tools["controller"].get_status()[Address])
            avetemp += temp
        temp = avetemp / self.num_temp_reads

        temp = self.apply_corrections(temp, temperature_id)

        # make nice float
        temp = float(int(temp * 1000.0) / 1000.0)

        # use some limits
        if temp > 100.0 or temp < -300.0:
            temp = -999.9

        # save temp
        self.last_temps[temperature_id] = temp

        return temp



import base64
import datetime
import io

from dash import html, dcc, callback, Input, Output, State, dash_table
import dash_bootstrap_components as dbc
import dash_daq as daq
import pandas as pd

import azcam
import azcam.web.queue.queue_ops as qops


def create_button_group():
    """
    Create button group for queue control.
    """

    button_group = dbc.ButtonGroup(
        [
            dbc.Button("Run script", id="run_btn", className="m-1", n_clicks=0),
            dbc.Button("Pause/Resume", id="pause_btn", className="m-1", n_clicks=0),
            dbc.Button("Abort script", id="abort_btn", className="m-1", n_clicks=0),
        ],
        vertical=False,
    )

    @callback(
        Output("hidden_div1", "children", allow_duplicate=True),
        Input("run_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_run(n):
        try:
            qops.execute()
        except Exception as e:
            print(e)
        return

    @callback(
        Output("hidden_div1", "children", allow_duplicate=True),
        Input("pausescript_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_pausescript(n):
        try:
            # azcam.db.tools["exposure"].reset()
            pass
        except Exception as e:
            print(e)
        return

    @callback(
        Output("hidden_div1", "children", allow_duplicate=True),
        Input("abortscript_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_abortscript(n):
        try:
            pass
            # azcam.db.tools["exposure"].abort()
        except Exception as e:
            print(e)
        return

    return button_group


def queue_card():
    """
    Create card for queue control.
    """

    cycles_input = html.Div(
        [
            dbc.Label("Cycles"),
            daq.NumericInput(id="cycles", value=1, size=100),
        ]
    )

    @callback(
        Output("hidden_div1", "children", allow_duplicate=True),
        Input("cycles", "value"),
        prevent_initial_call=True,
    )
    def cycles_callback(value):
        return

    altaz_input = html.Div(
        [
            dbc.Label("Use Alt/Az"),
            dbc.Checkbox(id="altaz", value=True),
        ]
    )

    @callback(
        Output("hidden_div1", "children", allow_duplicate=True),
        Input("altaz", "value"),
        prevent_initial_call=True,
    )
    def altaz_callback(value):
        return

    script_input = html.Div(
        [
            dcc.Upload(
                id="upload-data",
                children=html.Div(["Drag and Drop or ", html.A("Select Script")]),
                style={
                    "width": "100%",
                    "height": "60px",
                    "lineHeight": "60px",
                    "borderWidth": "1px",
                    "borderStyle": "dashed",
                    "borderRadius": "5px",
                    "textAlign": "center",
                    "margin": "10px",
                },
                multiple=False,
            ),
        ]
    )

    def parse_script(contents, filename):
        content_type, content_string = contents.split(",")

        decoded = base64.b64decode(content_string)
        try:
            if "csv" in filename:
                # Assume that the user uploaded a CSV file
                df = pd.read_csv(
                    io.StringIO(decoded.decode("utf-8")),
                    dtype=str,
                    keep_default_na=False,
                    skipinitialspace=True,
                )
                azcam.db.queue_df = df
        except Exception as e:
            print(e)
            return html.Div(["There was an error processing this file."])

        return html.Div(
            [
                html.H5(filename),
                dash_table.DataTable(
                    df.to_dict("records"), [{"name": i, "id": i} for i in df.columns]
                ),
            ]
        )

    @callback(
        Output("output-data-upload", "children"),
        Input("upload-data", "contents"),
        State("upload-data", "filename"),
        # State("upload-data", "last_modified"),
    )
    def update_output(list_of_contents, filename):

        if list_of_contents is not None:
            children = [parse_script(list_of_contents, filename)]
            return children

    button_group = create_button_group()

    queue_card = dbc.Card(
        [
            dbc.CardHeader("Observing Queue Control", style={"font-weight": "bold"}),
            dbc.CardBody(
                [
                    dbc.Row(
                        [
                            button_group,
                        ]
                    ),
                    dbc.Row(
                        [
                            dbc.Col(script_input),
                        ]
                    ),
                    dbc.Row(
                        [
                            dbc.Col(cycles_input),
                            dbc.Col(altaz_input),
                        ],
                    ),
                    dbc.Row(
                        [
                            html.Div(id="hidden_div1"),
                        ],
                    ),
                    dbc.Row(
                        [
                            html.Div(id="message_queue"),
                        ],
                    ),
                ]
            ),
            dbc.CardFooter(
                [
                    html.Div(id="output-data-upload"),
                ]
            ),
        ]
    )

    return queue_card



"""
Contains the Ds9Display class.
"""

"""
Note: This is a special tool as it is used by server and clients and so does not use 
the base Tools class.
"""

import os
import shutil
import subprocess
import tempfile
import time
from typing import List

import numpy
from astropy.io import fits as pyfits

import azcam
import azcam.utils
import azcam.exceptions
from azcam.tools.display import Display


class Ds9Display(Display):
    """
    Azcam's interface to SAO's ds9 image display tool.
    """

    def __init__(self):
        #: name used to reference the tool ("controller", "display", ...)
        self.tool_id: str = "display"

        #: descriptive tool name
        self.description: str = "ds9 display"

        super().__init__(self.tool_id, self.description)

        #: 1 when tool is enabled
        self.is_enabled: int = 1

        #: 1 when tool has been initialized
        self.is_initialized: int = 0

        #: 1 when tool has been reset
        self.is_reset: int = 0

        azcam.db.tools[self.tool_id] = self

        #: verbosity for debug, >0 is more verbose
        self.verbosity = 0

        # display Host, as a string hex code
        self.host = "0"
        # display Port, as a string decimal code
        self.port = "0"

        # image size for binary images
        self.size_x = 0
        self.size_y = 0

        # ROI's[first_col,last_col,first_row,last_row]
        # image coords     - entire image, binned units (starts at each amps reference)
        self.image_roi = []
        self.detector_roi = []  # detector coords  - single detector, unbinned units
        self.amp_roi = []  # amplifier coords - single amp, unbinned units
        self.return_integers = True  # true to return all coords as closest integers
        self.coordinate_type = "detector"

        # mean of current display ROI
        self.mean = 0.0
        # standard deviation of current display ROI
        self.sdev = 0.0

        # get names based on OS type
        self.root = None
        self.xpaproc = None

        # set default display server
        self.default_display = 0

    def initialize(self):
        """
        Initialize Ds9.

        :return None:
        """

        if self.is_initialized:
            return

        if os.name == "posix":
            self.xpaset_app = "xpaset"
            self.xpaget_app = "xpaget"
            self.xpaaccess_app = "xpaaccess"
            self.xpans = "xpans"
        else:
            self.root = "c:\\ds9\\"
            # self.root = "c:\\SAOImageDS9\\"
            self.ds9_app = os.path.join(self.root, "ds9.exe")
            self.ds9_app = os.path.abspath(self.ds9_app)
            self.xpaset_app = os.path.join(self.root, "xpaset.exe")
            self.xpaset_app = os.path.abspath(self.xpaset_app)
            self.xpaget_app = os.path.join(self.root, "xpaget.exe")
            self.xpaget_app = os.path.abspath(self.xpaget_app)
            self.xpaaccess_app = os.path.join(self.root, "xpaaccess.exe")
            self.xpaaccess_app = os.path.abspath(self.xpaaccess_app)
            self.xpans = os.path.join(self.root, "xpans.exe")
            self.xpans = os.path.abspath(self.xpans)

        if not self.is_enabled:
            azcam.exceptions.warning("Display is not enabled")
            return

        self.set_display(self.default_display)

        self.is_initialized = 1

        return

    def reset(self) -> None:
        """
        Reset the tool.
        """

        self.is_reset = 1

        return

    # *************************************************************************************************
    #   display server info
    # *************************************************************************************************
    def find_displays(self):
        """
        Returns a list of ds9 servers.
        The returned list is like: ['a00009c:50064', 'a00009c:59734'], which is host:port.
        If not servers are found, the returned list is empty.

        :return list: Display info from XPA
        """

        # command: xpaacess -v ds9

        resp = []

        cmd = self.xpaaccess_app
        p = subprocess.Popen(
            [self.xpaaccess_app, "-v", "ds9"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=False,
        )
        output, errors = p.communicate()
        output = output.decode("utf-8")
        if output == "":
            return []
        output = output.strip()
        resp = output.split("\n")

        return resp

    def set_display(self, display_number=-1):
        """
        Set the current ds9 display by number.

        :param int display_number: Number for display to be used (0->N)
        :return None:
        """

        if display_number == -1:
            display_number = self.default_display

        reply = self.find_displays()
        if not reply:
            return

        cnt = len(reply)  # number of displays available

        addrList = reply
        if cnt > 0 and display_number < cnt:
            if len(addrList) > 0:
                addr = addrList[display_number].split(":")
                self.host = addr[0]
                self.port = addr[1]
                self.default_display = display_number
        elif cnt > 0:  # one display, set reset default to it
            addr = addrList[0].split(":")
            self.host = addr[0]
            self.port = addr[1]
            self.default_display = 0
            return
        else:
            return

        return

    # *************************************************************************************************
    #   display image
    # *************************************************************************************************
    def display(self, image, extension_number=-1):
        """
        Display a file in ds9, making a copy so locking does not occur.
        If specified for an MEF file, only extension_number is displayed.

        :param image: a filename or an image object
        :param int extension_number: FITS extension number of image, -1 for all
        :return None:
        """

        self.initialize()

        # test, could be slow but seems to work nicely
        self.set_display()

        ds9 = self.host + ":" + self.port

        if type(image) == str:
            filename1 = azcam.utils.make_image_filename(image)
        else:
            filename1 = image.filename

        ext = os.path.splitext(filename1)[-1]

        # copy image file so it is not locked by ds9
        filename = os.path.join(tempfile.gettempdir(), "tempdisplayfile" + ext)
        shutil.copyfile(filename1, filename)

        if ext == ".fits":
            im = pyfits.open(filename)
            ne = max(0, len(im) - 1)
            im.close()
            if ne in [0, 1]:
                # s = self.xpaset_app + " " + ds9 + "fits iraf < " + filename
                s = [self.xpaset_app, ds9, f"fits iraf < {filename}"]
            else:
                if extension_number == -1:
                    # s = self.xpaset_app + " " + ds9 + "fits mosaicimage iraf < " + filename
                    s = [self.xpaset_app, ds9, f"fits mosaicimage iraf < {filename}"]
                else:
                    s = [
                        self.xpaset_app,
                        ds9,
                        "fits",
                        f"[{extension_number}] < {filename}",
                    ]
        elif ext == ".bin":
            NumCols, NumRows = self.size_x, self.size_y
            s = [
                self.xpaset_app,
                ds9,
                "array",
                f"[xdim={NumCols},ydim={NumRows},bitpix=-16] < {filename}",
            ]
        else:
            azcam.exceptions.warning("invalid image extension")
            return

        # execute XPA command to display file
        s = " ".join(s)
        p = subprocess.Popen(
            s,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        stdout, stderr = p.communicate()
        stdout = stdout.decode("utf-8")

        return

    # *************************************************************************************************

    #   ROI's
    # *************************************************************************************************
    def get_rois(self, roi_number=-1, coordinate_type="default"):
        """
        Returns the Region Of Interest from the display tool and sets the internal ROI descriptors
        to the values read.
        If roi_number is not specified then a list of all defined ROI's is returned.

        :param int roi_number: roi number to be returned
        :param str CoorinateType: Type of coords as defined by the display server ['image','detector', 'amplifier']
        :return list: The list of ROIs
        """

        if coordinate_type == "default":
            coordinate_type = self.coordinate_type

        # single ROI
        if roi_number != -1:
            roi = self._get_ds9_rois(coordinate_type)
            if not roi:
                return []
            if roi_number + 1 > len(roi):
                return []
            if self.return_integers:
                roi1 = []
                for r in roi[roi_number]:
                    roi1.append(int(r))
                return roi1
            else:
                return roi

        roi = self._get_ds9_rois(coordinate_type)
        if not roi:
            return []
        if not self.return_integers:
            return roi

        # make all ROIs integers
        roi1 = []
        for x in roi:
            roiN = []
            for r in x:
                roiN.append(int(r))
            roi1.append(roiN)

        return roi1

    def _get_ds9_rois(self, coordinate_type=""):
        """
        Returns all regions as a list of lists, each as [firstcol,lastcol,firstrow,lastrow].
        Ds9 regions are converted to [firstcol,lastcol,firstrow,lastrow] as close as possible.

        :param str coordinate_type: Type of coords as defined by the display server ['image','detector', 'amplifier']
        :return list: List of ROI's
        """

        if coordinate_type == "":
            coordinate_type = self.coordinate_type

        Rois = self.get_regions(
            coordinate_type
        )  # this is a list of roi's, each [shape, roi]
        if not Rois:
            return []

        rois = []
        for roi in Rois:
            if roi[0] == "box":  # box is the only shape supported at this time
                xcenter = roi[1]
                ycenter = roi[2]
                width = roi[3]
                length = roi[4]
            else:
                return []

            firstcol = xcenter - width / 2.0
            lastcol = xcenter + width / 2.0
            firstrow = ycenter - length / 2.0
            lastrow = ycenter + length / 2.0
            r = [firstcol, lastcol, firstrow, lastrow]
            rois.append(r)

        return rois

    def get_regions(self, coordinate_type="image"):
        """
        Returns a list of regions, each a list of [shape,coords...].

        :param str coordinate_type: Type of coords as defined by the display server ['image','detector', 'amplifier']
        :param list: List of ROI info in ds9 format
        """

        data = []

        # test, could be slow but seems to work nicely
        self.set_display()

        try:
            if coordinate_type == "amplifier":
                cmd = "regions -system amplifier -strip yes"
            elif coordinate_type == "detector":
                cmd = "regions -system detector -strip yes"
            elif coordinate_type == "image":
                cmd = "regions -system image -strip yes"
            else:
                return []

            datads9 = self.xpaget(cmd)
            datads9 = datads9.lstrip("detector;")
            datads9 = datads9.lstrip("image;")
            datads9 = datads9.lstrip("amplifier;")
            datads9 = datads9.rstrip(";")
            datads9 = datads9.split(";")
            # a list of regions
            if len(datads9) > 0:
                shape = "box"
                for roinum, d in enumerate(datads9):  # each region
                    d = d.lstrip("box(")
                    d = d.rstrip(")")
                    coords = d.split(",")
                    a = [float(x) for x in coords]
                    a.insert(0, shape)
                    data.append(a)
            else:
                return []

        except Exception as e:
            azcam.log(e)
            return []

        return data

    def read_rois(self):
        """
        Read and internally save image, detector, and amplifier all Rois.

        :return None:
        """

        self.image_roi = self.get_rois(-1, "image")
        if not self.image_roi:
            self.detector_roi = []
            self.amp_roi = []
            return
        self.detector_roi = self.get_rois(-1, "detector")
        self.amp_roi = self.get_rois(-1, "amplifier")

        return

    def get_number_rois(self):
        """
        Returns the number of defined ROI's.

        :return int: Number of ROI's specified by ds9
        """

        self.read_rois()
        return len(self.image_roi)

    def get_roi_string(self, roi_number=0, coordinate_type=""):
        """
        Get an ROI descriptor formatted as a string with integer values.

        :param int roi_number: roi_number to return
        :param str coordinate_type:

        :return list: List of ROIs
        """

        self.read_rois()
        if coordinate_type == "":
            coordinate_type = self.coordinate_type

        if roi_number > len(self.image_roi) - 1:
            return []

        if coordinate_type == "amplifier":
            roi = self.amp_roi[roi_number]
        elif coordinate_type == "detector":
            roi = self.detector_roi[roi_number]
        elif coordinate_type == "image":
            roi = self.image_roi[roi_number]
        else:
            return ""

        return " ".join(str(x) for x in roi)

    def get_last_point(self, coordinate_mode="physical"):
        """
        Returns coordinates of the last point selected.

        :param str coordinate_mode: Coorinate mode, wcs or pysical
        :return list: Coords of point as [x,y] or [] if none
        """
        posX = 0
        posY = 0
        # coordinate_mode = ''

        if coordinate_mode == "wcs":
            cmd = "regions -system wcs -strip yes"
        else:
            cmd = "regions -system physical -strip yes"

        datads9 = self.xpaget(cmd)
        if len(datads9) > 0:
            data = datads9[0]
            if coordinate_mode == "wcs":
                pts = data.lstrip("fk5")
            else:
                pts = data.lstrip("physical")
            pts = pts.split(";")

            for pt in pts:
                pt.strip()
                if len(pt) == 0:
                    continue
                if len(pt) > 0:
                    coord = pt.lstrip("point")
                    coord = pt.lstrip("box")
                    try:
                        coord = coord.lstrip("(")
                        coord = coord.rstrip(")")
                        nums = coord.split(",")
                        posX = nums[0]
                        posY = nums[1]
                    except Exception:
                        pass

            return [float(posX), float(posY)]

        else:
            return []

    def delete_all_rois(self):
        """
        Delete all ROIs.

        :return None:
        """

        cmd = "regions delete all"
        self.xpaset(cmd)

        return

    def draw_box(self, xy, size=[51, 51], angle=0):
        """
        Draw a box on the display.

        :param list xy: [x,y] coords of box to be drawn in image coordinates
        :param list size: [xwidth,ywidth] of box
        :param float angle: Angle of box.
        :return None:
        """

        cmd = f"regions command '{{box {xy[0]} {xy[1]} {size[0]} {size[1]} {angle}}}'"
        self.xpaset(cmd)

        return

    def get_box(self) -> List:
        """
        Get position of the first box in image coords.

        :return None:
        """

        self.read_rois()

        if len(self.image_roi) > 0:
            return self.image_roi[0]
        else:
            return []

    def set_cursor_mode(self, mode="point", clear_rois=0) -> None:
        """
        Set the display cursor mode.

        :param str mode: Cursor mode ("point", "crosshair")
        :param bool clear_rois: True to clear all ROIs
        """

        if mode == "point":
            self._set_pointer(clear_rois)
        elif mode == "crosshair":
            self._set_crosshair(clear_rois)
        else:
            raise azcam.exceptions.AzcamError("Invalid cursor mode")

        return

    def zoom(self, scale=0):
        """
        Set display zoom factor relative to current zoom.

        :param int Scale: Scale factor for zoom, 0 "to fit"
        :return None:
        """

        s = "to fit" if scale == 0 else str(scale)
        try:
            cmd = "zoom " + s
            self.xpaset(cmd)
        except Exception:
            return

    def _set_crosshair(self, clear_rois=0):
        """
        Set the ds9 pointer to a crosshair.

        :param bool clear_rois: True to clean all ROIs
        :return None:
        """

        if clear_rois:
            self.delete_all_rois()

        # set crosschair
        try:
            cmd = "mode crosshair"
            self.xpaset(cmd)
        except Exception:
            return

        return

    def _set_pointer(self, clear_rois=0):
        """
        Set the ds9 pointer to a pointer.

        :param bool clear_rois: True to clean all ROIs
        :return None:
        """

        # delete all reagions if clear = 1
        if clear_rois:
            self.delete_all_rois()

        # set pointer
        try:
            cmd = "mode pointer"
            self.xpaset(cmd)
        except Exception:
            return

        return

    # *************************************************************************************************
    #   examine displayed image
    # *************************************************************************************************
    def exam(self, box_size=20, loop=1):
        """
        Interactive: Get info around blinking cursor position.
        print() is allowed here as interactive only.

        :param int box_size: Size of examine box
        :param bool Loop: True for keep looping until *q* pressed
        :return list: [KeyPressed,X,Y] or [] in case or error.  KeyPressed is a character.
        """

        if loop:
            print("Press q to quit...")

        while 1:
            try:
                xsize = int(box_size)
                ysize = xsize
                cmd = "imexam key data %d %d" % (xsize, ysize)
                datads9 = self.xpaget(cmd)
                key = datads9[0]
                datads9 = datads9[1:]
                data = [float(d) for d in datads9]
                if key == "q":
                    loop = 0
                if loop:
                    print(
                        "key=%1s, mean=%7.0f, std=%6.01f"
                        % (key, float(numpy.mean(data)), float(numpy.std(data)))
                    )
                else:
                    return [key, numpy.mean(data), numpy.std(data)]
            except Exception:
                return []

            if not loop:
                break

    def get_stats(self, roi_number=-1):
        """
        Returns statistics [status,mean,sdev] of an  ROI
        Also sets display.mean and display.sdev global variables.

        :param int roi_number: Number of ROI to get status
        :return list: [status,mean,sdev] for ROI
        """

        data = self.get_data(roi_number)
        self.mean = numpy.mean(data)
        self.sdev = numpy.std(data)

        return [self.mean, self.sdev]

    def show_stats(self, roi_number=0):
        """
        Interactive: Print stats from an ROI in a real-time loop until *q* is pressed.
        print() is allowed here as interactive only.

        :param int roi_number: Number of ROI to get status
        :return list: [status,mean,sdev] for ROI
        """

        print("Displaying mean, sdev, ROI - press q to quit...")

        while 1:
            try:
                mean, sdev = self.get_stats(roi_number)
                print(
                    "mean: %.1f   sdev: %.2f   ROI: %s"
                    % (mean, sdev, str(self.get_roi_string(roi_number)))
                )
                key = azcam.utils.check_keyboard(0)
                if key == "q":
                    break
                time.sleep(0.1)
            except Exception:
                pass

            if azcam.utils.check_keyboard(0) == "q":
                break

        return

    def get_data(self, roi_number=0):
        """
        Returns [Status,[pixel1,pixel2,...]] in display ROI.

        :param int roi_number: Number of ROI to get data
        :return list: [status,[pixel1,pixel2,...]] for ROI
        """

        self.read_rois()
        numrois = len(self.detector_roi)
        if numrois == 0:
            raise azcam.exceptions.AzcamError("No ROI defined")
        elif roi_number > numrois:
            raise azcam.exceptions.AzcamError("Invalid ROI number")

        roi = self.detector_roi[roi_number]

        try:
            firstcol = roi[0]
            lastcol = roi[1]
            firstrow = roi[2]
            lastrow = roi[3]
            width = lastcol - firstcol + 1
            length = lastrow - firstrow + 1
            xcenter = firstcol + width / 2.0
            ycenter = firstrow + length / 2.0
            cmd = "data detector %d %d %d %d yes" % (
                xcenter,
                ycenter,
                width,
                length,
            )  # may not be right
            datads9 = self.xpaget(cmd)
            data = []
            datads9 = datads9.split("\n")
            for d in datads9:
                if len(d) == 0:
                    continue
                data.append(float(d))
            return data
        except Exception as e:
            azcam.log(f"ds9 error: {e}")
            return []

    # *************************************************************************************************
    #   save displayed image
    # *************************************************************************************************
    def save_image(self, filename="display.png"):
        """
        Save displayed image as a PNG snapshot.

        :param str filename: filename of image to be saved.
        :return None:
        """

        # many kluges here, XPA seems buggy on this.

        # problem with paths so use only filename
        # fullname=os.path.abspath(filename)
        # filename=os.path.basename(filename)
        # folder=os.path.dirname(fullname)
        fullname = filename

        f = fullname.lstrip("C:")

        # save image
        cmd = "saveimage png " + f
        self.xpaset(cmd)

        return None

    def save_fits(self, filename="display.fits"):
        """
        Save displayed image as a FITS file.

        :param str filename: filename of image to be saved.
        :return None:
        """

        folder = os.getcwd()
        filename = os.path.join(folder, filename)

        filename = os.path.normpath(filename)

        cmd = "save " + filename
        return self.xpaset(cmd)

    # *************************************************************************************************
    #   XPA commands
    # *************************************************************************************************
    def xpaget(self, command):
        """
        Issue xpaget command for ds9.
        Returns data as a list of space deliminated items.

        :param str command: command string for xpaget command
        :return: Return value
        """

        ds9 = self.host + ":" + self.port + " "

        cmd = self.xpaget_app + " " + ds9 + command

        # execute XPA command
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        stdout, stderr = p.communicate()
        stdout = stdout.decode("utf-8")

        return stdout

    def xpaset(self, command) -> None:
        """
        Issue xpaset command for ds9.

        :param str command: command string for xpaset command

        """

        ds9 = self.host + ":" + self.port + " "
        cmd = self.xpaset_app + " -p " + ds9 + command

        cmd = cmd.replace("\\\\", "/")
        cmd = cmd.replace("/", "\\")

        # execute XPA command
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        stdout, stderr = p.communicate()

        return



"""
*azcam.cmdserver* contains the CommandServer class for azcam's socket command interface.
"""

import os
import socket
import socketserver
import threading
import time
import json
from typing import Callable

import azcam
import azcam.utils
import azcam.exceptions


class CommandServer(socketserver.ThreadingTCPServer):
    """
    Main class for cmdserver.

    CommandServer class to receive and execute client commands over the socket interface.
    This is a socket server which receives command strings, executes them, and returns a reply string.
    The server normally runs in a thread so as to not block the command line. Each client which
    connects runs in its own thread (through the ThreadingTCPServer class) and so operates
    concurrently.  There is no global locking, so this process can be dangerous but it does allows
    multiple clients to interact simultaneously with azcam, which is important for operations
    like camera control, telescope movement, temperature readback, instrument control, etc.
    """

    def __init__(self, port=2402):
        self.welcome_message = None

        self.port = port  # listen port
        self.is_running = 0  # True when server is running
        self.server = 0  # server instance
        self.verbose = 0
        self.log_connections = 1  # log open and close connections
        self.monitorinterface = 0

        self.logcommands = 0

        self.socketnames = {}
        self.use_clientname = 1  # log client name with command

        self.currentclient = 0

        self.case_insensitive = 0

        azcam.db.cmdserver = self

    def begin(self, port=-1):
        """
        Start command server.
        """

        if port == -1:
            port = self.port
        else:
            self.port = port

        server_address = ("", port)  # '' better than localhost when no network

        try:
            self.server = ThreadedTCPServer(server_address, MyBaseRequestHandler)
            self.server.RequestHandlerClass.cmdserver = self

            self.is_running = 1
            self.server.serve_forever()  # waits here forever
        except Exception as message:
            self.is_running = 0
            azcam.log(
                f"ERROR in cmdserver:{repr(message)} Is it already running? Exiting..."
            )
            time.sleep(2)
            os._exit(1)

        # Exits here when server is aborted

        return

    def start(self, port=-1):
        """
        Starts command server in a thread.
        """

        cmdthread = threading.Thread(target=self.begin, name="cmdserver")
        cmdthread.daemon = True  # terminates when main process exits
        cmdthread.start()

        return

    def stop(self, port=-1):
        """
        Stops command server.
        """

        raise NotImplementedError

    def command(self, command: str):
        """
        Execute a command string received from a client over the command socket.
        Returns the reply string, always starting with OK or ERROR.
        """

        toolid, args, kwargs = self.parse_command_string(command, self.case_insensitive)

        reply = self.execute_command(toolid, args, kwargs)

        return reply

    def execute_command(self, tool: Callable, args: list, kwargs: dict = {}) -> str:
        """
        Executes a tool command which has been parsed into tool method and arguments and
        returns its reply string.

        Args:
            tool: tool object (not name)
            args: list of arguments
            kwargs: dictionary of keyword:value pairs for arguments

        Returns:
            reply: reply from command executed. Always starts with OK or ERROR.
        """

        if len(args) == 0 and len(kwargs) == 0:
            reply = tool()

        elif len(kwargs) == 0:
            reply = tool(*args)

        elif len(args) == 0:
            reply = tool(**kwargs)

        else:
            reply = tool(*args, **kwargs)

        reply = self._command_reply(reply)

        return reply

    def _command_reply(self, reply: str):
        """
        Create a reply string for a socket command.

        Args:
            reply (str): command reply

        Returns:
            [type]: formatted reply string
        """

        if reply is None or reply == "":
            s = ""

        elif type(reply) == str:
            s = reply

        elif type(reply) == list:
            s = ""
            for x in reply:
                if type(x) == str and " " in x:  # check if space in the string
                    s = s + " " + "'" + str(x) + "'"
                else:
                    s = s + " " + str(x)
            s = s.strip()

        elif type(reply) == dict:
            s = json.dumps(reply)

        else:
            s = repr(reply)

            if s != '""':
                s = s.strip()

        # add OK status if needed
        if not (s.startswith("OK") or s.startswith("ERROR") or s.startswith("WARNING")):
            s = "OK " + s

        s = s.strip()

        return s

    def parse_command_string(self, command: str, case_insensitive: int = 0):
        """
        Parse a command string into tool and arguments.
        If command does not start with a dotted object.method token, then
        assume it is the method of the default_tool.

        Returns (objid, args, kwargs)
        objid is a bound method of a class
        args is a list of strings
        kwargs is a dict of strings
        """

        # parse command string
        tokens = azcam.utils.parse(command, 0)
        cmd = tokens[0]

        if case_insensitive:
            cmd = cmd.lower()

        arglist = tokens[1:]
        args = []
        kwargs = {}
        if len(arglist) == 0:
            pass
        else:
            for token in arglist:
                if "=" in token:
                    keyname, value = token.split("=")
                    kwargs[keyname] = value
                else:
                    args.append(token)

        if "." not in cmd:
            # get method from db.default_tool
            if azcam.db.default_tool == "api":
                objid = getattr(azcam.db.api, cmd)
            elif azcam.db.default_tool is None:
                s = f"command not recognized: {cmd} "
                raise azcam.exceptions.AzcamError(s)
            else:
                objid = getattr(azcam.db.tools[azcam.db.default_tool], cmd)

        else:
            # get method from tool in db.tools
            objects = cmd.split(".")

            # special case temporarily for parameters
            if objects[0] == "parameters":
                if len(objects) == 1:
                    objid = azcam.db.parameters
                elif len(objects) == 2:
                    objid = getattr(azcam.db.parameters, objects[1])
                elif len(objects) == 3:
                    objid = getattr(
                        getattr(azcam.db.parameters, objects[1]), objects[2]
                    )
                elif len(objects) == 4:
                    objid = getattr(
                        getattr(getattr(azcam.db.parameters, objects[1]), objects[2]),
                        objects[3],
                    )
                else:
                    objid = None  # too complicated for now
                return objid, args, kwargs

            elif objects[0] == "api":
                objid = getattr(azcam.db.api, objects[1])
                return objid, args, kwargs

            elif objects[0] not in azcam.db.tools:
                raise azcam.exceptions.AzcamError(
                    f"remote call not allowed: {objects[0]}", 4
                )

            if len(objects) == 1:
                objid = azcam.db.tools[objects[0]]
            elif len(objects) == 2:
                objid = getattr(azcam.db.tools[objects[0]], objects[1])
            elif len(objects) == 3:
                objid = getattr(
                    getattr(azcam.db.tools[objects[0]], objects[1]), objects[2]
                )
            elif len(objects) == 4:
                objid = getattr(
                    getattr(
                        getattr(azcam.db.tools[objects[0]], objects[1]), objects[2]
                    ),
                    objects[3],
                )
            else:
                objid = None  # too complicated for now

            # kwargs = {}
            # l1 = len(tokens)
            # if l1 > 1:
            #     args = tokens[1:]
            #     if "=" in args[0]:
            #         # assume all keywords for now
            #         kwargs = {}
            #         for argtoken in args:
            #             keyword, value = argtoken.split("=")
            #             kwargs[keyword] = value
            #         args = []
            # else:
            #     args = []

        return objid, args, kwargs


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    # allow_reuse_address = True
    allow_reuse_address = False


class MyBaseRequestHandler(socketserver.BaseRequestHandler):
    def __init__(self, request, client_address, server):
        azcam.db.cmdserver.currentclient += 1
        self.currentclient = azcam.db.cmdserver.currentclient

        socketserver.BaseRequestHandler.__init__(self, request, client_address, server)

    def handle(self):
        """
        Called when a connection is made from a client.
        Starts an infinite loop waiting for new commands.
        Commands are executed sequentially.
        """

        if azcam.db.cmdserver.welcome_message is not None:
            self.request.send(str.encode(azcam.db.cmdserver.welcome_message + "\r\n"))

        while True:
            try:
                prefix_in = f"Rcv{self.currentclient:01}> "
                prefix_out = f"Out{self.currentclient:01}>  "  # extra space for indent

                # ************************************************************************
                # receive command from the network socket
                # ************************************************************************
                try:
                    command_string = self.receive_command()
                except ConnectionResetError:
                    azcam.log(
                        f"Client {azcam.db.tools['cmdserver'].socketnames[self.currentclient]} disconnected",
                        prefix=prefix_in,
                    )
                    break
                except Exception as e:
                    azcam.log(f"ERROR in handle: {e}", prefix="Err-> ")
                    break

                # ************************************************************************
                # disconnect on empty string - important
                # ************************************************************************
                if command_string.strip() == "":
                    try:
                        self.request.send(str.encode("OK\r\n"))
                    except OSError:
                        pass
                    except Exception as e:
                        azcam.log(
                            f"Null command send error for client {self.currentclient}: {e}"
                        )
                    break

                # ************************************************************************
                # log received command
                # ************************************************************************
                try:
                    self.cmdserver.socketnames[self.currentclient]
                except Exception:
                    azcam.db.cmdserver.socketnames[self.currentclient] = (
                        f"unknown_{self.currentclient}"
                    )
                if azcam.db.cmdserver.logcommands:
                    azcam.log(command_string, prefix=prefix_in)

                # ************************************************************************
                # check special cases which do not leave cmdserver
                # ************************************************************************

                # close socket connection to client
                if command_string.lower().startswith("closeconnection"):
                    azcam.log(
                        f"closing connection to {azcam.db.tools['cmdserver'].socketnames[self.currentclient]}",
                        prefix=prefix_in,
                    )
                    self.request.send(str.encode("OK\r\n"))
                    self.request.close()
                    break

                # register - register a client name, example: register console
                elif command_string.lower().startswith("register"):
                    x = command_string.split(" ")
                    azcam.db.cmdserver.socketnames[self.currentclient] = (
                        f"{x[1]}_{int(self.currentclient)}"
                    )
                    self.request.send(str.encode("OK\r\n"))
                    azcam.log(
                        f"OK client {self.currentclient}", prefix=prefix_out
                    )  # log reply
                    command_string = ""

                # echo - for polling as "echo hello" or just "echo"
                elif command_string.lower().startswith("echo"):
                    s = command_string.split(" ")
                    if len(s) == 1:
                        reply = "OK"
                    elif len(s) == 2:
                        reply = "OK %s" % s[1]
                    else:
                        reply = "OK %s" % " ".join(s[1:])
                    self.request.send(str.encode(reply + "\r\n"))
                    if azcam.db.cmdserver.logcommands:
                        azcam.log("%s" % reply, prefix=prefix_out)
                    command_string = ""

                # update - azcammonitor
                elif command_string.lower().startswith("updatemonitor"):
                    azcam.db.monitor.register()
                    azcam.log("%s" % "OK", prefix=prefix_out)
                    reply = "OK"

                    self.request.send(str.encode(reply + "\r\n"))
                    command_string = ""

                # exit - send reply for handshake before closing socket and shutting down
                elif command_string.lower().startswith("exit"):
                    self.request.send(str.encode("OK\r\n"))
                    azcam.log("%s" % "OK", prefix=prefix_out)  # log reply
                    self.request.close()
                    os._exit(0)  # kill python

                # ************************************************************************
                # process all other command_strings
                # ************************************************************************
                if command_string != "":
                    # execute command
                    try:
                        reply = azcam.db.cmdserver.command(command_string)
                    except Exception as e:
                        reply = f"ERROR {repr(e)}"

                    # log reply
                    if azcam.db.cmdserver.logcommands:
                        azcam.log(reply, prefix=prefix_out)

                    # send reply to socket
                    self.request.send(str.encode(reply + "\r\n"))

                else:
                    time.sleep(0.10)  # for telnet

            except Exception as message:  # catch everything so cmdserver never crashes
                azcam.log(f"ERROR in cmdserver: {command_string}: {message}")
                # try to reply but this may not work
                try:
                    self.request.send(str.encode(f"ERROR {repr(message)}\r\n"))
                except Exception as e:
                    print(e)
                    pass  # OK to do nothing

        return

    def setup(self):
        """
        Called when new connection made.
        """

        if azcam.db.cmdserver.log_connections and azcam.db.cmdserver.verbose:
            azcam.log(
                f"Client connection made from {str(self.client_address)}",
                prefix="cmd> ",
            )

        return socketserver.BaseRequestHandler.setup(self)

    def finish(self):
        """
        Called when existing connection is closed.
        """

        if azcam.db.cmdserver.log_connections and azcam.db.cmdserver.verbose:
            azcam.log(f"Connection closed to {str(self.client_address)}")

        return socketserver.BaseRequestHandler.finish(self)

    def receive_command(self):
        """
        Receive a string from socket until terminator is found.
        Returns a string.
        Returns empty string on error.
        """

        terminator = "\n"  # likely ends with \r\n

        # read socket until terminator found
        msg = ""
        msg1 = ""
        while True:
            try:
                msg1 = self.request.recv(1024).decode()
                if msg1 == "":
                    return ""
                if msg1[-1] == terminator:  # found terminator
                    msg += msg1[:-1]
                    break
                msg += msg1
            except socket.error as e:
                if e.errno == 10054:  # connection closed
                    pass
                else:
                    azcam.log(f"receive_command: {e}", prefix="Err-> ")
                break

        if len(msg) == 0:
            msg = ""

        if msg is None:
            msg = ""

        return msg



"""
Configure and start fastapi application using uvicorn.
Import this after all configuration has been completed.
All API commands suported here must start with ""http://locahost:xxxx/api/" where xxxx is the
port number like 2402.

Query string example: "http://localhost:2402/api/set_filter?filter=1&filter_id=2"

JSON example:
    data = {
        "command": "set_par",
        "args": [],
        "kwargs": {"parameter": "imagetest", "value": 3333},
    }
    r = requests.post("http://localhost:2403/api", json=data1)
    print(r.status_code, r.json())
    
Default response is JSON:
    response = {
        "message": "Finished",
        "command": urlparse(url).path,
        "data": reply,
    }

"""

import os
import threading

import uvicorn
from fastapi import FastAPI, Request, APIRouter, HTTPException
from fastapi.responses import FileResponse
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.templating import Jinja2Templates
from fastapi.middleware.wsgi import WSGIMiddleware

from azcam.web.exposure.exposure_web import ExposureWeb
from azcam.web.queue.queue_web import QueueWeb
from azcam.web.status.status_web import StatusWeb


import azcam
from azcam import exceptions


class WebServer(object):
    """
    Azcam web server.
    """

    def __init__(self):
        self.templates_folder = ""
        self.index = "index.html"
        self.favicon_path = None

        self.logcommands = 0
        self.logstatus = 0
        self.message = ""  # customized message
        self.datafolder = None

        # port for webserver
        self.port = None

        self.is_running = 0

        azcam.db.webserver = self

    def initialize(self):
        """
        Initialize application.
        """

        # create app
        app = FastAPI()
        self.app = app

        if self.datafolder is None:
            self.datafolder = os.path.dirname(__file__)

        if self.favicon_path is None:
            self.favicon_path = os.path.join(os.path.dirname(__file__), "favicon.ico")

        self.index = os.path.join(os.path.dirname(__file__), self.index)
        self.version = 1  # API version

        # Plotly Dash webs
        expweb = ExposureWeb()
        expweb.logcommands = 1
        expweb.logstatus = 0
        queueweb = QueueWeb()
        queueweb.logcommands = 1
        queueweb.logstatus = 0
        statusweb = StatusWeb()
        statusweb.logcommands = 1
        statusweb.logstatus = 0
        app.mount("/exposure", WSGIMiddleware(expweb.app.server))
        app.mount("/queue", WSGIMiddleware(queueweb.app.server))
        app.mount("/status", WSGIMiddleware(statusweb.app.server))

        # templates folder
        try:
            templates = Jinja2Templates(directory=os.path.dirname(self.index))
        except Exception:
            pass

        # ******************************************************************************
        # Home - /
        # ******************************************************************************
        @app.get("/", response_class=HTMLResponse)
        def home(request: Request):
            index = os.path.basename(self.index)
            return templates.TemplateResponse(
                index,
                {
                    "request": request,
                    "message": self.message,
                    "webport": azcam.db.cmdserver.port + 1,
                },
            )

        # ******************************************************************************
        # Set API version: /api/set_version?version=2
        # ******************************************************************************
        @app.get("/api/set_version", response_class=JSONResponse)
        async def set_version(request: Request, command: str = None):

            qpars = request.query_params
            self.version = int(qpars["version"])

            response = {
                "message": "Finished",
                "command": f"api.set_version",
                "data": self.version,
            }

            return JSONResponse(response)

        @app.get("/favicon.ico", include_in_schema=False)
        async def favicon():
            return FileResponse(self.favicon_path)

        # ******************************************************************************
        # Example of a special case: /api/exposure/get_status
        # ******************************************************************************
        @app.get("/api/exposure/get_status", response_class=JSONResponse)
        async def expstatus(request: Request, command: str = None):

            expstatus = azcam.db.api.get_status()

            response = {
                "message": "Finished",
                "command": f"exposure.get_status",
                "data": expstatus,
            }

            return JSONResponse(response)

        # ******************************************************************************
        # API interface - ../api/command or ../api JSON
        # ******************************************************************************

        @app.post("/api", response_class=JSONResponse)
        @app.get("/api/{command:path}", response_class=JSONResponse)
        async def api(request: Request, command: str = None):
            """
            Remote web api commands.
            """

            if self.version == 1 or self.version == 2:

                # JSON
                if command is None:
                    args = await request.json()
                    try:
                        toolid = azcam.db.api
                        command = getattr(toolid, args["command"])

                        arglist = args["args"]
                        kwargs = args["kwargs"]
                        reply = command(*arglist, **kwargs)
                    except Exception as e:
                        reply = repr(e)
                        azcam.log(e)

                    response = {
                        "message": "Finished",
                        "command": f"api.{args['command']}",
                        "data": reply,
                    }

                    return JSONResponse(response)

                # query string
                url = command
                qpars = request.query_params

                if self.logcommands:
                    azcam.log(url, prefix="Web-> ")

                reply = self.web_command(url, qpars)

                if self.logcommands:
                    azcam.log(reply, prefix="---->   ")

                return JSONResponse(reply)

            else:

                # query string
                url = command
                qpars = request.query_params

                response = {
                    "message": "Finished",
                    "command": url,
                    "data": "Invalid API version",
                }

                return JSONResponse(response)

    # ******************************************************************************
    # webserver methods
    # ******************************************************************************

    def add_router(self, router):
        """
        Add router.
        """

        self.app.include_router(router)

        return

    def stop(self):
        """
        Stops command server running in thread.
        """

        azcam.log("Stopping the webserver is not supported")

        return

    def start(self):
        """
        Start web server.
        """

        self.initialize()

        if self.port is None:
            self.port = azcam.db.cmdserver.port + 1

        azcam.log(f"Starting webserver - listening on port {self.port}")

        # uvicorn.run(self.app)

        arglist = [self.app]
        kwargs = {"port": self.port, "host": "0.0.0.0", "log_level": "critical"}

        thread = threading.Thread(
            target=uvicorn.run, name="uvicorn", args=arglist, kwargs=kwargs
        )
        thread.daemon = True  # terminates when main process exits
        thread.start()

        self.is_running = 1

        return

    def web_command(self, url, qpars=None):
        """
        Parse and execute a command string received as a URL.
        Returns the reply as a JSON packet.
        """

        try:
            method = url
            kwargs = qpars._dict

            caller = getattr(azcam.db.api, method)
            reply = caller() if kwargs is None else caller(**kwargs)

        except Exception as e:
            azcam.log(e)
            reply = f"invalid API command: {url}"

        # generic response
        response = {
            "message": "Finished",
            "command": url,
            "data": reply,
        }

        return response



"""
Contains the base TempConCryoCon24 class.
"""

import socket
import threading

import azcam
import azcam.exceptions
from azcam.tools.tempcon import TempCon

"""
Example commands:
    tempcon.server.command("input A:units C")
    tempcon.server.command("input B:units C")
    tempcon.server.command("input C:units C")
    tempcon.server.command("input A:isenix 2")  # LS DT-670
    tempcon.server.command("input B:isenix 2")  # LS DT-670
    tempcon.server.command("input C:isenix 2")  # LS DT-670
    tempcon.server.command("loop 1:type pid")  # PID control
    tempcon.server.command("loop 1:range mid")  # range (0.5W,5W,50W)
    tempcon.server.command("loop 1:maxpwr 100")  # max power on this range
    tempcon.server.command("loop 1:source B")  # controlling input channel
"""


class TempConCryoCon24(TempCon):
    """
    Cryogenic Control Systems Model 24 temperature control class.
    """

    def __init__(self, tool_id="tempcon", description=None):
        super().__init__(tool_id, description)

        self.host = ""
        self.port = 5000

        self.temperature_ids = [0, 1]  # camtemp, dewtemp

        self.temp_names = {0: "A", 1: "B", 2: "C", 3: "D"}

        self.init_commands = []

    def define_keywords(self):
        """
        Defines and resets instrument keywords.
        """

        self.set_keyword("CAMTEMP", 0.0, "Camera temperature", "float")
        self.set_keyword("DEWTEMP", 0.0, "Dewar temperature", "float")
        self.set_keyword("TEMPUNIT", "C", "Temperature units", "str")

        return

    def initialize(self):
        """
        Initialize CryoConM24TemperatureControl temperature controller.
        """

        if self.is_initialized:
            return
        if not self.is_enabled:
            azcam.exceptions.warning(f"{self.description} is not enabled")
            return

        # recreate in case host/port changed
        self.server = TempconServerInterface(self.host, self.port, self.description)

        # query ID
        try:
            self.server.command("*IDN?;")
        except socket.timeout as e:
            azcam.exceptions.warning("Could not query temperature controller: {e}")
            return

        # set PID mode
        self.server.command("LOOP 1:TYPE PID")

        # set PID mode
        self.server.command("LOOP 2:TYPE OFF")

        # set control temp
        self.set_control_temperature(self.control_temperature)

        # run custom initialization commands
        for cmd in self.init_commands:
            self.server.command(cmd)

        self.is_initialized = 1

        return

    def set_control_temperature(
        self, temperature: float = None, temperature_id: int = -1
    ):
        """
        Set Loop 1 control temperature in Celsius.
        Args:
            temperature: temperature to set
            temperature_id: temperature ID number
                * 0 is TempA
                * 1 is TempB
                * 2 is TempC
                * 3 is TempD
        """

        if temperature is None:
            temperature = self.control_temperature
        else:
            self.control_temperature = float(temperature)

        temperature_id = int(temperature_id)

        if temperature_id != -1:
            self.control_temperature_id = temperature_id

        # set loop 1
        self.server.command(f"Loop 1:SETPT {temperature};")
        self.server.command(f"LOOP 1:TYPE PID")
        self.server.command(
            f"LOOP 1:SOURCE {self.temp_names[self.control_temperature_id]}"
        )

        # turn on control loop
        self.server.command("CONTrol;")

        return

    def get_temperature(self, temperature_id: int = 0):
        """
        Reads temperatures from the tempcon controller.
        Args:
            temperature_id: temperature ID number
        Returns:
            temperature: temperature read
        """

        if not self.is_enabled:
            # azcam.exceptions.warning("Tempcon not enabled")
            return -999.9

        if not self.is_initialized:
            # azcam.exceptions.warning("Tempcon not initialized")
            return -999.9

        temperature_id = int(temperature_id)
        if temperature_id == 0:
            tempstr = "INPUT? A;"
        elif temperature_id == 1:
            tempstr = "INPUT? B;"
        elif temperature_id == 2:
            tempstr = "INPUT? C;"
        elif temperature_id == 3:
            tempstr = "INPUT? D;"
        else:
            raise azcam.exceptions.AzcamError("bad temperature_id in get_temperature")

        reply = self.server.command(tempstr)
        try:
            temp = float(reply)
        except ValueError:
            temp = self.bad_temp_value

        temp = self.apply_corrections(temp, temperature_id)

        return temp


class TempconServerInterface(object):
    """
    Communicates with the CryoCon Model 24 builtin temperature server using a socket.
    """

    def __init__(self, host: str, port: int, name: str):
        self.host = host
        self.port = port
        self.connected = 0
        self.cmdstring = ""
        self.timeout = 5.0

        self.lock = threading.Lock()

    def open(self):
        """
        Open a socket connection to an server.
        Creates the socket and makes a connection.
        """

        if not self.connected:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.socket.settimeout(self.timeout)

        return

    def close(self):
        """
        Close an open socket connection to an server.
        """

        self.socket.close()
        self.connected = 0

        return

    def command(self, command: str):
        """
        Communicate with the remote server.
        Sends and command and returns reply.
        Returns the exact reply from the server.
        Args:
            command: command string to send.
        Returns:
            reply: commands reply
        """

        with self.lock:
            self.open()
            self.socket.sendto(
                str.encode(command), (self.host, self.port + 1)
            )  # UDP is TCP port + 1
            reply = self.socket.recvfrom(4000)
            self.close()
            reply = reply[0].decode()  # UDP return string is first element
            reply = reply[:-2]  # remove CR/LF

        return reply
